# Copyright (C) 2010, 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.
#
from django.core.files.base import ContentFile
from django.db import models, transaction, IntegrityError

import logging


class BundleManager(models.Manager):

    @transaction.commit_on_success
    def create_with_content(self, bundle_stream, uploaded_by, content_filename, content):
        logging.debug("Creating bundle object")
        bundle = self.create(
                bundle_stream=bundle_stream,
                uploaded_by=uploaded_by,
                content_filename=content_filename)
        # XXX: this _can_ fail -- if content_sha1 is a duplicate
        logging.debug("Saving bundle object (this is safe so far)")
        bundle.save()
        try:
            logging.debug("saving bundle content (file) and bundle object")
            bundle.content.save("bundle-{0}".format(bundle.pk),
                                ContentFile(content))
        except IntegrityError as exc:
            logging.debug("integrity error: %r", exc)
            # Note: we're not saving the deletion back to the database
            # because we are going to rollback anyway. In PostgreSQL this
            # would also always fail because the database is not going to
            # honor any other operations until we rollback.
            # See: 
            # http://docs.djangoproject.com/en/1.2/topics/db/transactions/#handling-exceptions-within-postgresql-transactions
            logging.debug("deleting content file")
            bundle.content.delete(save=False)
            raise
        else:
            return bundle 



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests of the TestClient support class
"""
from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test import TestCase

from dashboard_app.tests.utils import TestClient


def auth_test(request):
    response = HttpResponse(mimetype="text/plain")
    if (request.user and request.user.is_authenticated and
        request.user.is_active):
        response.write(request.user.username)
    response['Content-length'] = str(len(response.content))
    return response


class local_urls:
    from django.conf.urls.defaults import patterns, url
    urlpatterns = patterns('', url(r'^auth-test/$', auth_test))


class TestClientTest(TestCase):

    _USER = "user"

    urls = local_urls

    def setUp(self):
        super(TestClientTest, self).setUp()
        self.client = TestClient()
        self.user = User(username=self._USER)
        self.user.save()

    def test_auth(self):
        self.client.login_user(self.user)
        response = self.client.get("/auth-test/")
        self.assertEqual(response.content, self._USER)

    def test_no_auth(self):
        response = self.client.get("/auth-test/")
        self.assertEqual(response.content, '')




# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for the BundleStream model
"""

from django.contrib.auth.models import User, Group
from django.db import IntegrityError
from django_testscenarios.ubertest import TestCaseWithScenarios

from dashboard_app.models import BundleStream


class BundleStreamTests(TestCaseWithScenarios):

    _NAME = "name"
    _SLUG = "slug"
    _GROUPNAME = "group"
    _USERNAME = "user"

    scenarios = [
        ('anonymous-no-slug', {
            'pathname': '/anonymous/',
            'is_public': "true",
            'is_anonymous': "true",
            'username': "user",
        }),
        ('anonymous-with-slug', {
            'name': _NAME,
            'slug': _SLUG,
            'is_public': "true",
            'is_anonymous': "true",
            'pathname': '/anonymous/slug/',
            'username': "user",
        }),
        ('personal-no-slug', {
            'username': _USERNAME,
            'pathname': '/private/personal/user/',
        }),
        ('personal-with-slug', {
            'username': _USERNAME,
            'name': _NAME,
            'slug': _SLUG,
            'pathname': '/private/personal/user/slug/',
        }),
        ('team-no-slug', {
            'groupname': _GROUPNAME,
            'pathname': '/private/team/group/',
        }),
        ('team-with-slug', {
            'groupname': _GROUPNAME,
            'name': _NAME,
            'slug': _SLUG,
            'pathname': '/private/team/group/slug/',
        }),
    ]

    groupname = None
    username = None
    group = None
    user = None
    name = ''
    slug = ''
    is_public = 0
    is_anonymous = 0

    def setUp(self):
        super(BundleStreamTests, self).setUp()
        if self.username is not None:
            self.user = User.objects.create(username='user')
        if self.groupname is not None:
            self.group = Group.objects.create(name='group')

    def test_creation(self):
        bundle_stream = BundleStream.objects.create(
            user=self.user, group=self.group, name=self.name, slug=self.slug)
        bundle_stream.save()
        self.assertEqual(bundle_stream.user, self.user)
        self.assertEqual(bundle_stream.group, self.group)
        self.assertEqual(bundle_stream.name, self.name)
        self.assertEqual(bundle_stream.slug, self.slug)

    def test_team_named_stream(self):
        bundle_stream = BundleStream.objects.create(
            user=self.user, group=self.group, name=self.name, slug=self.slug, 
            is_anonymous=self.is_anonymous, is_public=self.is_public)
        bundle_stream.save()
        self.assertEqual(bundle_stream.pathname, self.pathname)

    def test_pathname_uniqueness(self):
        bundle_stream = BundleStream.objects.create(
            user=self.user, group=self.group, name=self.name, slug=self.slug)
        bundle_stream.save()
        self.assertRaises(
            IntegrityError,
            BundleStream.objects.create,
            user=self.user, group=self.group, slug=self.slug,
            name=self.name
        )

    def test_pathname_update(self):
        bundle_stream = BundleStream.objects.create(
            user=self.user, group=self.group, name=self.name, slug=self.slug)
        bundle_stream.save()
        old_pathname = bundle_stream.pathname
        bundle_stream.slug += "-changed"
        bundle_stream.save()
        self.assertNotEqual(bundle_stream.pathname, old_pathname)
        self.assertEqual(
            bundle_stream.pathname,
            bundle_stream._calc_pathname()
        )

    def test_unicode(self):
        obj = BundleStream(pathname=self.pathname)
        self.assertEqual(unicode(obj), self.pathname)



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Administration interface of the Dashboard application
"""

from django import forms
from django.contrib import admin
from django.contrib.contenttypes import generic
from django.utils.translation import ugettext as _

from dashboard_app.models import (
        Attachment,
        Bundle,
        BundleDeserializationError,
        BundleStream,
        HardwareDevice,
        NamedAttribute,
        SoftwarePackage,
        SoftwareSource,
        Test,
        TestCase,
        TestResult,
        TestRun,
        )


class BundleAdmin(admin.ModelAdmin):

    def bundle_stream_pathname(self, bundle):
        return bundle.bundle_stream.pathname
    bundle_stream_pathname.short_description = _("Bundle stream")

    list_display = ('bundle_stream_pathname', 'content_filename',
            'uploaded_by', 'uploaded_on', 'is_deserialized')
    list_filter = ('bundle_stream',)
    readonly_fields = ('is_deserialized',)
    date_hierarchy = 'uploaded_on'
    fieldsets = (
        ('Document', {
            'fields': ('content', 'content_filename')}),
        ('Upload Details', {
            'fields': ('bundle_stream', 'uploaded_by')}),
        ('Deserialization', {
            'fields': ('is_deserialized',)}),
    )


class BundleDeserializationErrorAdmin(admin.ModelAdmin):
    list_display = ('bundle', 'error_message')
    search_fields = ('bundle__content_sha1',)


class BundleStreamAdminForm(forms.ModelForm):
    class Meta:
        model = BundleStream

    def clean(self):
        cleaned_data = self.cleaned_data
        if (cleaned_data.get('user', '') is not None and
                cleaned_data.get('group') is not None):
            raise forms.ValidationError('BundleStream cannot have both user '
                    'and name set at the same time')
        return super(BundleStreamAdminForm, self).clean()


class BundleStreamAdmin(admin.ModelAdmin):
    form = BundleStreamAdminForm
    list_display = ('pathname', 'user', 'group', 'slug', 'is_public', 'is_anonymous', 'name')
    list_filter = ('is_public', 'is_anonymous')
    prepopulated_fields = {"slug": ("name",)}
    readonly_fields = ('pathname',)
    fieldsets = (
            ('Name and description', {
                'fields': ('name', 'slug', 'pathname')}),
            ('Ownership', {
                'fields': ('user', 'group')}),
            ('Access Rights', {
                'fields': ('is_public', 'is_anonymous')}),
            )


class SoftwarePackageAdmin(admin.ModelAdmin):
    list_display = ('name', 'version')
    search_fields = ('name', 'version')


class SoftwareSourceAdmin(admin.ModelAdmin):
    list_display = ('project_name', 'branch_url', 'branch_vcs',
                    'branch_revision', 'commit_timestamp')
    list_filter = ('project_name',)
    search_fields = ('project_name', 'branch_url')


class HardwareDeviceAdmin(admin.ModelAdmin):
    class NamedAttributeInline(generic.GenericTabularInline):
        model = NamedAttribute
    list_display = ('description', 'device_type')
    list_filter = ('device_type',)
    search_fields = ('description',)
    inlines = [NamedAttributeInline]


class TestCaseAdmin(admin.ModelAdmin):
    list_display = ('test_case_id', 'test',)
    list_filter = ('test',)


class TestAdmin(admin.ModelAdmin):
    pass


class TestResultAdmin(admin.ModelAdmin):
    class NamedAttributeInline(generic.GenericTabularInline):
        model = NamedAttribute
    list_display = ('__unicode__', 'test', 'test_case', 'result', 'measurement')
    list_filter = ('test_case', 'result')
    inlines = [NamedAttributeInline]


class TestRunAdmin(admin.ModelAdmin):
    class NamedAttributeInline(generic.GenericTabularInline):
        model = NamedAttribute
    list_filter = ('test'),
    list_display = (
        'test',
        'analyzer_assigned_uuid',
        'bundle',
        'analyzer_assigned_date',
        'import_assigned_date')
    inlines = [NamedAttributeInline]


admin.site.register(Attachment)
admin.site.register(Bundle, BundleAdmin)
admin.site.register(BundleDeserializationError, BundleDeserializationErrorAdmin)
admin.site.register(BundleStream, BundleStreamAdmin)
admin.site.register(HardwareDevice, HardwareDeviceAdmin)
admin.site.register(SoftwarePackage, SoftwarePackageAdmin)
admin.site.register(SoftwareSource, SoftwareSourceAdmin)
admin.site.register(Test, TestAdmin)
admin.site.register(TestCase, TestCaseAdmin)
admin.site.register(TestResult, TestResultAdmin)
admin.site.register(TestRun, TestRunAdmin)



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Dashboard Application (package)
"""

__version__ = (0, 7, 1, "final", 0)



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

from django.core.urlresolvers import reverse
from django_testscenarios.ubertest import TestCase

from dashboard_app.tests import fixtures


class RedirectTests(TestCase):

    _PATHNAME = "/anonymous/"
    _BUNDLE_TEXT = """
{
  "test_runs": [
    {
      "test_results": [
        {
          "test_case_id": "test-case-0", 
          "result": "pass"
        } 
      ], 
      "analyzer_assigned_date": "2010-10-15T22:04:46Z", 
      "time_check_performed": false, 
      "analyzer_assigned_uuid": "00000000-0000-0000-0000-000000000001",
      "test_id": "examples"
    }
  ], 
  "format": "Dashboard Bundle Format 1.0"
}
    """
    _BUNDLE_NAME = "whatever.json"

    def setUp(self):
        super(RedirectTests, self).setUp()
        self.bundle = fixtures.create_bundle(self._PATHNAME, self._BUNDLE_TEXT, self._BUNDLE_NAME)
        self.bundle.deserialize()
        self.assertTrue(self.bundle.is_deserialized)

    def test_bundle_permalink(self):
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_bundle",
                    args=(self.bundle.content_sha1, )))
        self.assertRedirects(response, self.bundle.get_absolute_url())

    def test_bundle_permalink_trailing(self):
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_bundle",
                    args=(self.bundle.content_sha1, 'trailing/')))
        self.assertRedirects(
            response, self.bundle.get_absolute_url() + 'trailing/',
            target_status_code=404)

    def test_bundle_permalink_query_string(self):
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_bundle",
                    args=(self.bundle.content_sha1, )), data={'foo': 'bar'})
        self.assertRedirects(
            response, self.bundle.get_absolute_url()+'?foo=bar')

    def test_test_run_permalink(self):
        test_run = self.bundle.test_runs.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_run",
                    args=(test_run.analyzer_assigned_uuid, )))
        self.assertRedirects(response, test_run.get_absolute_url())

    def test_test_run_permalink_trailing(self):
        test_run = self.bundle.test_runs.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_run",
                    args=(test_run.analyzer_assigned_uuid, 'trailing/')))
        self.assertRedirects(
            response, test_run.get_absolute_url() + 'trailing/',
            target_status_code=404)

    def test_test_run_permalink_query_string(self):
        test_run = self.bundle.test_runs.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_run",
                    args=(test_run.analyzer_assigned_uuid, )),
            data={'foo': 'bar'})
        self.assertRedirects(
            response, test_run.get_absolute_url() + '?foo=bar')

    def test_test_result_permalink(self):
        test_run = self.bundle.test_runs.all()[0]
        test_result = test_run.test_results.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_result",
                    args=(test_run.analyzer_assigned_uuid,
                          test_result.relative_index)))
        self.assertRedirects(response, test_result.get_absolute_url())

    def test_test_result_permalink_trailing(self):
        test_run = self.bundle.test_runs.all()[0]
        test_result = test_run.test_results.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_result",
                    args=(test_run.analyzer_assigned_uuid,
                          test_result.relative_index, 'trailing/')))
        self.assertRedirects(
            response, test_result.get_absolute_url() + 'trailing/',
            target_status_code=404)

    def test_test_result_permalink_query_string(self):
        test_run = self.bundle.test_runs.all()[0]
        test_result = test_run.test_results.all()[0]
        response = self.client.get(
            reverse("dashboard_app.views.redirect_to_test_result",
                    args=(test_run.analyzer_assigned_uuid,
                          test_result.relative_index)),
            data={'foo': 'bar'})
        self.assertRedirects(
            response, test_result.get_absolute_url() + '?foo=bar')



from dashboard_app.models import ImageHealth


def dashboard_globals(request):
    return {
        "dashboard": {
            "interesting_images": [
                ImageHealth('nano', 'panda'),
                ImageHealth('nano', 'omap3'),
                ImageHealth('nano', 'lt-panda'),
                ImageHealth('developer', 'panda'),
                ImageHealth('developer', 'omap3'),
                ImageHealth('developer', 'lt-panda'),
                ImageHealth('alip', 'panda'),
                ImageHealth('alip', 'omap3'),
                ImageHealth('alip', 'lt-panda'),
                ImageHealth('alip', 'lt-panda-x11-base-natty'),
                ImageHealth('ubuntu-desktop', 'panda'),
                ImageHealth('ubuntu-desktop', 'omap3'),
                ImageHealth('ubuntu-desktop', 'lt-panda'),
                ImageHealth('ubuntu-desktop', 'lt-panda-x11-base-natty'),
            ]
        }
    }



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Regression test for LP:658917
"""

from django.db import IntegrityError

from django_testscenarios.ubertest import TestCase

from dashboard_app.models import Bundle
from dashboard_app.xmlrpc import DashboardAPI
from pkg_resources import resource_string

from dashboard_app.tests import fixtures

class LP658917(TestCase):


    def setUp(self):
        super(LP658917, self).setUp()
        self.bundle_stream = fixtures.create_bundle_stream("/anonymous/")
        self.dashboard_api = DashboardAPI()
        self.content_sha1 = None

    def tearDown(self):
        if self.content_sha1:
            Bundle.objects.get(content_sha1=self.content_sha1).delete_files()
        super(LP658917, self).tearDown()

    def test_658917(self):
        """TestCase.units is not assigned a null value"""
        try:
            self.content_sha1 = self.dashboard_api.put(
                resource_string(__name__, 'LP658917.json'),
                'LP658917.json', self.bundle_stream.pathname)
        except IntegrityError:
            self.fail("LP658917 regression, IntegrityError raised")



# Copyright (C) 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
DataViews: Encapsulated SQL query definitions.

Implementation of the following launchpad blueprint:
https://blueprints.launchpad.net/launch-control/+spec/other-linaro-n-data-views-for-launch-control
"""


from xml.sax import parseString

from dashboard_app.repositories import Repository, Undefined, Object
from dashboard_app.repositories.common import BaseContentHandler


class _DataViewHandler(BaseContentHandler):

    """
    ContentHandler subclass for parsing DataView documents
    """
    
    def startDocument(self):
        # Classic-classes 
        BaseContentHandler.startDocument(self)
        # Data view object
        self.obj = Object()
        # Set default values
        self.obj.name = Undefined
        self.obj.backend_queries = {}
        self.obj.arguments = []
        self.obj.documentation = None
        self.obj.summary = None
        # Internal variables
        self._current_backend_query = None

    def endDocument(self):
        # TODO: check if we have anything defined
        if self.obj.name is Undefined:
            raise ValueError("No data view definition found")

    def startElement(self, name, attrs):
        if name == "data-view":
            self.obj.name = attrs["name"]
        elif name == "summary" or name == "documentation":
            self._start_text()
        elif name == "sql":
            self._start_text()
            self._current_backend_query = BackendSpecificQuery(
                attrs.get("backend"), None, [])
            self.obj.backend_queries[
                self._current_backend_query.backend] = self._current_backend_query
        elif name == "value":
            if "name" not in attrs:
                raise ValueError("<value> requires attribute 'name'")
            self._text.append("{" + attrs["name"] + "}")
            self._current_backend_query.argument_list.append(attrs["name"])
        elif name == "argument":
            if "name" not in attrs:
                raise ValueError("<argument> requires attribute 'name'")
            if "type" not in attrs:
                raise ValueError("<argument> requires attribute 'type'")
            if attrs["type"] not in ("string", "number", "boolean", "timestamp"):
                raise ValueError("invalid value for argument 'type' on <argument>")
            argument = Argument(name=attrs["name"], type=attrs["type"],
                                   default=attrs.get("default", None),
                                   help=attrs.get("help", None))
            self.obj.arguments.append(argument)
                
    def endElement(self, name):
        if name == "sql":
            self._current_backend_query.sql_template = self._end_text()
            self._current_backend_query = None
        elif name == "documentation":
            self.obj.documentation = self._end_text()
        elif name == "summary":
            self.obj.summary = self._end_text()


class Argument(object):
    """
    Data view argument for SQL prepared statements
    """

    def __init__(self, name, type, default, help):
        self.name = name
        self.type = type
        self.default = default
        self.help = help 


class BackendSpecificQuery(object):
    """
    Backend-specific query and argument list
    """

    def __init__(self, backend, sql_template, argument_list):
        self.backend = backend
        self.sql_template = sql_template
        self.argument_list = argument_list


class DataViewRepository(Repository):

    @property
    def settings_variable(self):
        return "DATAVIEW_DIRS"

    def load_from_xml_string(self, text):
        handler = _DataViewHandler()
        parseString(text, handler)
        return self.item_cls(**handler.obj.__dict__)


__all__ = [
    "DataViewRepository",
]



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for the Attachment model
"""
from django.contrib.contenttypes import generic
from django.core.files.base import ContentFile
from django.db import models
from django.test import TestCase

from dashboard_app.models import Attachment


class ModelWithAttachments(models.Model):
    """
    Test model that uses attachments
    """
    attachments = generic.GenericRelation(Attachment)

    class Meta:
        # This requires a bit of explanation. Traditionally we could add new
        # models inside test modules and they would be picked up by django and
        # synchronized (created in the test database) as a part of the test
        # provisioning process.

        # Since we started using south, synchronization is no longer occurring
        # for the 'dashboard_app' application. This caused some test failures
        # such as any tests that depended on the existence of this model.

        # As a workaround we artificially "stick" this model into the only
        # application that we can count on to exist _and_ not use south as well
        # -- that is south itself.

        # This way the test model gets synchronized when south is synchronized
        # and all the test code below works as expected.
        app_label = "south"


class AttachmentTestCase(TestCase):
    _CONTENT = "text"
    _FILENAME = "filename"

    def setUp(self):
        self.obj = ModelWithAttachments.objects.create()

    def test_attachment_can_be_added_to_models(self):
        attachment = self.obj.attachments.create(
            content_filename = self._FILENAME, content=None)
        self.assertEqual(attachment.content_object, self.obj)

    def test_attachment_can_be_accessed_via_model(self):
        self.obj.attachments.create(
            content_filename = self._FILENAME, content=None)
        self.assertEqual(self.obj.attachments.count(), 1)
        retrieved_attachment = self.obj.attachments.all()[0]
        self.assertEqual(retrieved_attachment.content_object, self.obj)

    def test_attachment_stores_data(self):
        attachment = self.obj.attachments.create(
            content_filename = self._FILENAME, content=None)
        attachment.content.save(
            self._FILENAME,
            ContentFile(self._CONTENT))
        self.assertEqual(attachment.content_filename, self._FILENAME)
        attachment.content.open()
        try:
            self.assertEqual(attachment.content.read(), self._CONTENT)
        finally:
            attachment.content.close()
            attachment.content.delete(save=False)

    def test_unicode(self):
        obj = Attachment(content_filename="test.json")
        self.assertEqual(unicode(obj), "test.json")



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
XMP-RPC API
"""

import decimal
import logging
import xmlrpclib

from django.contrib.auth.models import User, Group
from django.core.urlresolvers import reverse
from django.db import IntegrityError, DatabaseError
from linaro_django_xmlrpc.models import (
    ExposedAPI,
    Mapper,
    xml_rpc_signature,
)

from dashboard_app import __version__
from dashboard_app.models import (
    Bundle,
    BundleStream,
    DataView,
)


class errors:
    """
    A namespace for error codes that may be returned by various XML-RPC
    methods. Where applicable existing status codes from HTTP protocol
    are reused
    """
    AUTH_FAILED = 100
    AUTH_BLOCKED = 101
    FORBIDDEN = 403
    NOT_FOUND = 404
    CONFLICT = 409
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501


class DashboardAPI(ExposedAPI):
    """
    Dashboard API object.

    All public methods are automatically exposed as XML-RPC methods
    """

    data_view_connection = DataView.get_connection()

    @xml_rpc_signature('str')
    def version(self):
        """
        Name
        ----
        `version` ()

        Description
        -----------
        Return dashboard server version. The version is a string with
        dots separating five components.

        The components are:
            1. major version
            2. minor version
            3. micro version
            4. release level
            5. serial

        See: http://docs.python.org/library/sys.html#sys.version_info

        Return value
        -------------
        Server version string
        """
        return ".".join(map(str, __version__))

    def _put(self, content, content_filename, pathname):
        try:
            logging.debug("Getting bundle stream")
            bundle_stream = BundleStream.objects.accessible_by_principal(self.user).get(pathname=pathname)
        except BundleStream.DoesNotExist:
            logging.debug("Bundle stream does not exists, aborting")
            raise xmlrpclib.Fault(errors.NOT_FOUND,
                    "Bundle stream not found")
        try:
            logging.debug("Creating bundle object")
            bundle = Bundle.objects.create_with_content(bundle_stream, self.user, content_filename, content)
        except (IntegrityError, ValueError) as exc:
            logging.debug("Raising xmlrpclib.Fault(errors.CONFLICT)")
            raise xmlrpclib.Fault(errors.CONFLICT, str(exc))
        except:
            logging.exception("big oops")
            raise
        else:   
            logging.debug("Deserializing bundle")
            bundle.deserialize()
            return bundle

    @xml_rpc_signature('str', 'str', 'str', 'str')
    def put(self, content, content_filename, pathname):
        """
        Name
        ----
        `put` (`content`, `content_filename`, `pathname`)

        Description
        -----------
        Upload a bundle to the server.

        Arguments
        ---------
        `content`: string
            Full text of the bundle. This *SHOULD* be a valid JSON
            document and it *SHOULD* match the "Dashboard Bundle Format
            1.0" schema. The SHA1 of the content *MUST* be unique or a
            ``Fault(409, "...")`` is raised. This is used to protect
            from simple duplicate submissions.
        `content_filename`: string
            Name of the file that contained the text of the bundle. The
            `content_filename` can be an arbitrary string and will be
            stored along with the content for reference.
        `pathname`: string
            Pathname of the bundle stream where a new bundle should
            be created and stored. This argument *MUST* designate a
            pre-existing bundle stream or a ``Fault(404, "...")`` exception
            is raised. In addition the user *MUST* have access
            permission to upload bundles there or a ``Fault(403, "...")``
            exception is raised. See below for access rules.

        Return value
        ------------
        If all goes well this function returns the SHA1 of the content.

        Exceptions raised
        -----------------
        404
            Either:

                - Bundle stream not found
                - Uploading to specified stream is not permitted
        409
            Duplicate bundle content

        Rules for bundle stream access
        ------------------------------
        The following rules govern bundle stream upload access rights:
            - all anonymous streams are accessible
            - personal streams are accessible to owners
            - team streams are accessible to team members

        """
        bundle = self._put(content, content_filename, pathname)
        logging.debug("Returning bundle SHA1")
        return bundle.content_sha1

    @xml_rpc_signature('str', 'str', 'str', 'str')
    def put_ex(self, content, content_filename, pathname):
        """
        Name
        ----
        `put` (`content`, `content_filename`, `pathname`)

        Description
        -----------
        Upload a bundle to the server.  A variant on put_ex that returns the
        URL of the bundle instead of its SHA1.

        Arguments
        ---------
        `content`: string
            Full text of the bundle. This *SHOULD* be a valid JSON
            document and it *SHOULD* match the "Dashboard Bundle Format
            1.0" schema. The SHA1 of the content *MUST* be unique or a
            ``Fault(409, "...")`` is raised. This is used to protect
            from simple duplicate submissions.
        `content_filename`: string
            Name of the file that contained the text of the bundle. The
            `content_filename` can be an arbitrary string and will be
            stored along with the content for reference.
        `pathname`: string
            Pathname of the bundle stream where a new bundle should
            be created and stored. This argument *MUST* designate a
            pre-existing bundle stream or a ``Fault(404, "...")`` exception
            is raised. In addition the user *MUST* have access
            permission to upload bundles there or a ``Fault(403, "...")``
            exception is raised. See below for access rules.

        Return value
        ------------
        If all goes well this function returns the full URL of the bundle.

        Exceptions raised
        -----------------
        404
            Either:

                - Bundle stream not found
                - Uploading to specified stream is not permitted
        409
            Duplicate bundle content

        Rules for bundle stream access
        ------------------------------
        The following rules govern bundle stream upload access rights:
            - all anonymous streams are accessible
            - personal streams are accessible to owners
            - team streams are accessible to team members

        """
        bundle = self._put(content, content_filename, pathname)
        logging.debug("Returning permalink to bundle")
        return self._context.request.build_absolute_uri(
            reverse(
                'dashboard_app.views.redirect_to_bundle',
                kwargs={'content_sha1':bundle.content_sha1}))

    def get(self, content_sha1):
        """
        Name
        ----
        `get` (`content_sha1`)

        Description
        -----------
        Download a bundle from the server.

        Arguments
        ---------
        `content_sha1`: string
            SHA1 hash of the content of the bundle to download. This
            *MUST* designate an bundle or ``Fault(404, "...")`` is raised.

        Return value
        ------------
        This function returns an XML-RPC struct with the following fields:

        `content_filename`: string
            The value that was stored on a previous call to put()
        `content`: string
            The full text of the bundle

        Exceptions raised
        -----------------
        404
            Either:

                - Bundle not found
                - Downloading from the stream that contains this bundle is
                  not permitted

        Rules for bundle stream access
        ------------------------------
        The following rules govern bundle stream download access rights:
            - all anonymous streams are accessible
            - personal streams are accessible to owners
            - team streams are accessible to team members
        """
        try:
            bundle = Bundle.objects.get(content_sha1=content_sha1)
            if not bundle.bundle_stream.is_accessible_by(self.user):
                raise Bundle.DoesNotExist()
        except Bundle.DoesNotExist:
            raise xmlrpclib.Fault(errors.NOT_FOUND,
                    "Bundle not found")
        else:
            return {"content": bundle.content.read(),
                    "content_filename": bundle.content_filename}

    @xml_rpc_signature('struct')
    def streams(self):
        """
        Name
        ----
        `streams` ()

        Description
        -----------
        List all bundle streams that the user has access to

        Arguments
        ---------
        None

        Return value
        ------------
        This function returns an XML-RPC array of XML-RPC structs with
        the following fields:

        `pathname`: string
            The pathname of the bundle stream
        `name`: string
            The user-configurable name of the bundle stream
        `user`: string
            The username of the owner of the bundle stream for personal
            streams or an empty string for public and team streams.
        `group`: string
            The name of the team that owsn the bundle stream for team
            streams or an empty string for public and personal streams.
        `bundle_count`: int
            Number of bundles that are in this stream

        Exceptions raised
        -----------------
        None

        Rules for bundle stream access
        ------------------------------
        The following rules govern bundle stream download access rights:
            - all anonymous streams are accessible
            - personal streams are accessible to owners
            - team streams are accessible to team members
        """
        bundle_streams = BundleStream.objects.accessible_by_principal(self.user)
        return [{
            'pathname': bundle_stream.pathname,
            'name': bundle_stream.name,
            'user': bundle_stream.user.username if bundle_stream.user else "",
            'group': bundle_stream.group.name if bundle_stream.group else "",
            'bundle_count': bundle_stream.bundles.count(),
            } for bundle_stream in bundle_streams]

    def bundles(self, pathname):
        """
        Name
        ----
        `bundles` (`pathname`)

        Description
        -----------
        List all bundles in a specified bundle stream

        Arguments
        ---------
        `pathname`: string
            The pathname of the bundle stream to query. This argument
            *MUST* designate an existing stream or Fault(404, "...") is
            raised. The user *MUST* have access to this stream or
            Fault(403, "...") is raised.

        Return value
        ------------
        This function returns an XML-RPC array of XML-RPC structs with
        the following fields:

        `uploaded_by`: string
            The username of the user that uploaded this bundle or 
            empty string if this bundle was uploaded anonymously.
        `uploaded_on`: datetime
            The timestamp when the bundle was uploaded
        `content_filename`: string
            The filename of the original bundle file
        `content_sha1`: string
            The SHA1 hash if the content of the bundle
        `content_size`: int
            This element was added in server version 0.4
            The size of the content
        `is_deserialized`: bool
            True if the bundle was de-serialized successfully, false otherwise


        Exceptions raised
        -----------------
        404
            Either:

                - Bundle stream not found
                - Listing bundles in this bundle stream is not permitted

        Rules for bundle stream access
        ------------------------------
        The following rules govern bundle stream download access rights:
            - all anonymous streams are accessible
            - personal streams are accessible to owners
            - team streams are accessible to team members
        """
        try:
            bundle_stream = BundleStream.objects.accessible_by_principal(self.user).get(pathname=pathname)
        except BundleStream.DoesNotExist:
            raise xmlrpclib.Fault(errors.NOT_FOUND, "Bundle stream not found")
        return [{
            'uploaded_by': bundle.uploaded_by.username if bundle.uploaded_by else "",
            'uploaded_on': bundle.uploaded_on,
            'content_filename': bundle.content_filename,
            'content_sha1': bundle.content_sha1,
            'content_size': bundle.content.size,
            'is_deserialized': bundle.is_deserialized
            } for bundle in bundle_stream.bundles.all().order_by("uploaded_on")]


    def deserialize(self, content_sha1):
        """
        Name
        ----
        `deserialize` (`content_sha1`)

        Description
        -----------
        Deserialize bundle on the server

        Arguments
        ---------
        `content_sha1`: string
            SHA1 hash of the content of the bundle to download. This
            *MUST* designate an bundle or ``Fault(404, "...")`` is raised.

        Return value
        ------------
        True - deserialization okay
        False - deserialization not needed

        Exceptions raised
        -----------------
        404
            Bundle not found
        409
            Bundle import failed
        """
        try:
            bundle = Bundle.objects.get(content_sha1=content_sha1)
        except Bundle.DoesNotExist:
            raise xmlrpclib.Fault(errors.NOT_FOUND, "Bundle not found")
        if bundle.is_deserialized:
            return False
        bundle.deserialize()
        if bundle.is_deserialized is False:
            raise xmlrpclib.Fault(
                errors.CONFLICT,
                bundle.deserialization_error.get().error_message)
        return True

    def make_stream(self, pathname, name):
        """
        Name
        ----
        `make_stream` (`pathname`, `name`)

        Description
        -----------
        Create a bundle stream with the specified pathname 

        Arguments
        ---------
        `pathname`: string
            The pathname must refer to an anonymous stream
        `name`: string
            The name of the stream (free form description text)

        Return value
        ------------
        pathname is returned

        Exceptions raised
        -----------------
        403
            Pathname does not designate an anonymous stream 
        409
            Bundle stream with the specified pathname already exists

        Available Since
        ---------------
        0.3
        """
        # Work around bug https://bugs.launchpad.net/lava-dashboard/+bug/771182
        # Older clients would send None as the name and this would trigger an
        # IntegrityError to be raised by BundleStream.objects.create() below
        # which in turn would be captured by the fault handler and reported as
        # an unrelated issue to the user. Let's work around that by using an
        # empty string instead.
        if name is None:
            name = ""
        try:
            user_name, group_name, slug, is_public, is_anonymous = BundleStream.parse_pathname(pathname)
        except ValueError as ex:
            raise xmlrpclib.Fault(errors.FORBIDDEN, str(ex))

        # Start with those to simplify the logic below
        user = None
        group = None
        if is_anonymous is False:
            if self.user is not None:
                assert is_anonymous is False
                assert self.user is not None
                if user_name is not None:
                    if user_name != self.user.username:
                        raise xmlrpclib.Fault(
                            errors.FORBIDDEN,
                            "Only user {user!r} could create this stream".format(user=user_name))
                    user = self.user  # map to real user object
                elif group_name is not None:
                    try:
                        group = self.user.groups.get(name=group_name)
                    except Group.DoesNotExist:
                        raise xmlrpclib.Fault(
                            errors.FORBIDDEN,
                            "Only a member of group {group!r} could create this stream".format(group=group_name))
            else:
                assert is_anonymous is False
                assert self.user is None
                raise xmlrpclib.Fault(
                    errors.FORBIDDEN, "Only anonymous streams can be constructed (you are not signed in)")
        else:
            assert is_anonymous is True
            assert user_name is None
            assert group_name is None
            # Hacky but will suffice for now
            user = User.objects.get_or_create(username="anonymous-owner")[0]
        try:
            bundle_stream = BundleStream.objects.create(
                user=user,
                group=group,
                slug=slug,
                is_public=is_public,
                is_anonymous=is_anonymous,
                name=name)
        except IntegrityError:
            raise xmlrpclib.Fault(
                errors.CONFLICT,
                "Stream with the specified pathname already exists")
        else:
            return bundle_stream.pathname

    def data_views(self):
        """
        Name
        ----
        `data_views` ()

        Description
        -----------
        List all data views

        Arguments
        ---------
        None

        Return value
        ------------
        This function returns an XML-RPC array of XML-RPC structs with
        the following fields:

        `name`: string
            Data view name declared in the definition file
        `summary`: string
            One-line description string suitable for developers

        Exceptions raised
        -----------------
        None
        """
        return [{
            'name': data_view.name,
            'summary': data_view.summary,
            "documentation": data_view.documentation,
            "arguments": [{
                "name": arg.name,
                "type": arg.type,
                "help": arg.help,
                "default": arg.default
            } for arg in data_view.arguments]
        } for data_view in DataView.repository.all()]

    def data_view_info(self, name):
        """
        Name
        ----
        `data_view_info` (`name`)

        Description
        -----------
        Describe a specific data view. This function looks up data view by name and returns rich information. See below for details

        Arguments
        ---------
        `name`: string
            Name of the data view to lookup
             

        Return value
        ------------
        This function returns an XML-RPC struct with the following fields:

        `name`: string
            Data view name declared in the definition file
        `summary`: string
            One-line description string suitable for developers
        `documentation`: string
            Longer documentation that described the purpose and indented usage of this data view
        `sql`: string or null
            The SQL of query specific to the currently running database (the
            actual query that is executed by query-data-view. Since some data
            views use database specific SQL the query may not be available.
        `argments` an XML-RPC array of XML-RPC structs with the following fields:
            `name`: Argument name
            `type`: Argument type, one of "number", "string" or "boolean"
            `help`: Help string for this argument
            `default`: Default value of an argument (or null if not available)


        Exceptions raised
        -----------------
        404
            Name does not designate a data view
        """
        try:
            data_view = DataView.repository.get(name=name)
        except DataView.DoesNotExist:
            raise xmlrpclib.Fault(errors.NOT_FOUND, "Data view not found")
        else:
            query = data_view.get_backend_specific_query(self.data_view_connection)
            return {
                "name": data_view.name,
                "summary": data_view.summary,
                "documentation": data_view.documentation,
                "sql": query.sql_template if query is not None else None, 
                "arguments": [{
                    "name": arg.name,
                    "type": arg.type,
                    "help": arg.help,
                    "default": arg.default
                } for arg in data_view.arguments]
            }

    def query_data_view(self, name, arguments):
        """
        Name
        ----
        `query_data_view` (name, arguments)

        Description
        -----------
        List all data views

        Arguments
        ---------
        None

        Return value
        ------------
        This function returns an XML-RPC struct with the following fields:

        `rows`: XML-RPC array of XML-RPC arrays
            Each item corresponds to cell in a row
        `columns`: XML-RPC array of XML-RPC structs with the following fields:
            `name`: XML-RPC string - name of the column
            `type`: XML-RPC string - column type (future extension, currently unused)

        Exceptions raised
        -----------------
        TBD
        """
        try:
            data_view = DataView.repository.get(name=name)
        except DataView.DoesNotExist:
            raise xmlrpclib.Fault(errors.NOT_FOUND, "Data view not found")
        try:
            rows, columns = data_view(self.data_view_connection, **arguments)
        except (LookupError, TypeError, ValueError, DatabaseError) as exc:
            raise xmlrpclib.Fault(errors.INTERNAL_SERVER_ERROR, str(exc))
        else:
            rows = [[float(cell) if isinstance(cell, decimal.Decimal) else cell for cell in row] for row in rows]
            return {
                "rows": rows,
                "columns": [{
                    "name": item[0],
                    "type": item[1]
                } for item in columns]
            }


# Mapper used by the legacy URL
legacy_mapper = Mapper()
legacy_mapper.register_introspection_methods()
legacy_mapper.register(DashboardAPI, '')



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for Cross-Site Request Forgery middleware configuration
"""
import xmlrpclib

import django
from django import forms
from django.conf.urls.defaults import patterns, url
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.template import Template, RequestContext

from dashboard_app.tests.utils import CSRFTestCase
from dashboard_app import urls


class CSRFConfigurationTestCase(CSRFTestCase):

    @property
    def urls(self):
        urlpatterns = urls.urlpatterns
        urlpatterns += patterns('', url(r'^test-form/', test_form))
        return type('urls', (), dict(urlpatterns=urlpatterns))

    def setUp(self):
        super(CSRFConfigurationTestCase, self).setUp()
        self.form_path = reverse(test_form)

    def test_csrf_token_present_in_form(self):
        if django.VERSION[:2] == (1, 1):
            # This feature is not supported on django 1.1
            return
        response = self.client.get(self.form_path)
        self.assertContains(response, "csrfmiddlewaretoken")

    def test_cross_site_form_submission_fails(self):
        if django.VERSION[:2] == (1, 1):
            # This feature is not supported on django 1.1
            return
        response = self.client.post(self.form_path, {'text': 'text'})
        self.assertEquals(response.status_code, 403)

    def test_csrf_not_protecting_xml_rpc_views(self):
        """call version and check that we didn't get 403"""
        endpoint_path = reverse('dashboard_app.views.dashboard_xml_rpc_handler')
        request_body = xmlrpclib.dumps((), methodname="version")
        response = self.client.post(endpoint_path, request_body, "text/xml")
        self.assertContains(response, "<methodResponse>", status_code=200)


def test_form(request):
    t = Template(template)
    html = t.render(RequestContext(request, {'form': SingleTextFieldForm()}))
    return HttpResponse(html)


class SingleTextFieldForm(forms.Form):
    text = forms.CharField()


template = """
    <html>
     <body>
      <form action="." method="POST">
      {% csrf_token %} 
       <table>{{ form.as_table }}</table>
      </form>
     </body>
    </html>
    """



# Copyright (C) 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

import abc
import logging
import os

from django.conf import settings


class RepositoryItemMeta(abc.ABCMeta):
    """
    Meta class for RepositoryItem

    Adds magic to set repository._item_cls to the non-meta
    class object that use this meta-class.
    """

    def __new__(mcls, name, bases, namespace):
        cls = super(RepositoryItemMeta, mcls).__new__(
            mcls, name, bases, namespace)
        if "repository" in namespace:
            repo = cls.repository
            repo.item_cls = cls
        return cls


class RepositoryItem(object):
    """
    Repository Item, element of a Repository.

    Each repository item is loaded from a XML file.
    """

    __metaclass__ = RepositoryItemMeta

    _base_path = None

    def _load_from_external_representation(self, pathname):
        self._base_path = os.path.dirname(pathname)

    @property
    def base_path(self):
        return self._base_path

    class DoesNotExist(Exception):
        pass

    class MultipleValuesReturned(Exception):
        pass


class RepositoryQuerySet(object):
    """
    QuerySet-like class for poking at RepositoryItems
    """

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def __nonzero__(self):
        return len(self._items) != 0

    def __getitem__(self, slice_or_index):
        return self._items.__getitem__(slice_or_index)

# Django QuerySet like API:

    def count(self):
        return len(self)

    def all(self):
        return self

    def get(self, **kwargs):
        query = self.filter(**kwargs)
        if len(query) == 1:
            return query[0]
        if not query:
            raise self.model.DoesNotExist()
        else:
            raise self.model.MultipleValuesReturned()

    def filter(self, **kwargs):
        return self._sub_query([
            item for item in self._items
            if all(
                (getattr(item, attr) == value)
                for attr, value in kwargs.iteritems())])

# Internal API

    def __init__(self, model, items):
        self.model = model
        self._items = items

    def _sub_query(self, items):
        return self.__class__(self.model, items)


class Repository(object):
    """
    Repository

    A container of XML documents loaded from disk that behave somewhat like
    django-built in database ORM
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.item_cls = None  # later patched by RepositoryItemMeta
        self._items = []
        self._did_load = False

    def _queryset(self):
        # In development mode always reload repository items
        if getattr(settings, "DEBUG", False) is True:
            self._did_load = False
        if not self._did_load:
            self._items = []
            self._load_default()
            self._did_load = True
        return RepositoryQuerySet(self.item_cls, self._items)

    def all(self):
        return self._queryset().all()

    def filter(self, **kwargs):
        return self._queryset().filter(**kwargs)

    def get(self, **kwargs):
        return self._queryset().get(**kwargs)

    def load_from_directory(self, directory):
        try:
            items = os.listdir(directory)
        except (OSError, IOError) as exc:
            logging.exception("Unable to enumreate directory: %s: %s",
                              directory, exc)
        else:
            for name in items:
                pathname = os.path.join(directory, name)
                if os.path.isfile(pathname) and pathname.endswith(".xml"):
                    self.load_from_file(pathname)

    @abc.abstractmethod
    def load_from_xml_string(self, text):
        """
        Load an IRepositoryItem from specified XML text
        """

    def load_from_file(self, pathname):
        try:
            with open(pathname, "rt") as stream:
                text = stream.read()
            item = self.load_from_xml_string(text)
            # Let the item know where it came from
            item._load_from_external_representation(pathname)
            self._items.append(item)
        except Exception as exc:
            logging.exception("Unable to load object into repository %s: %s",
                              pathname, exc)

    @abc.abstractproperty
    def settings_variable(self):
        """
        VARIABLE to look for in django settings. It should
        contain a list of directories to look for .xml
        files.
        """

    def _load_default(self):
        from django.conf import settings
        for dirname in getattr(settings, self.settings_variable, []):
            self.load_from_directory(dirname)


class Undefined(object):
    """
    Undefined object, as in JavaScript.

    Similar to None but serves different purpose. While None is often "empty"
    Undefined literally means "not set", this allows to use None as a normal
    (non-special cased value).
    """
    def __repr__(self):
        return "undefined"

# Undefined object singleton
Undefined = Undefined()


class Object(object):
    """
    Object, as in JavaScript.

    The only noticeable difference from plain python object is that undefined
    attributes do not raise AttributeError and instead produce Undefined
    values.
    """

    def __getattr__(self, name):
        return Undefined



"""
Test for the TestCase model
"""

from django.db import IntegrityError
from django.test import TestCase
from django_testscenarios.ubertest import TestCaseWithScenarios

from dashboard_app.models import (
    Test,
    TestCase as TestCaseModel,
)


class TestCaseConstructionTests(TestCaseWithScenarios):

    scenarios = [
        ('simple1', {
            'test_id': 'org.linaro.testheads.android',
            'test_case_id': 'testcase1',
            'name': "Boot test",
            'units': '',
        }),
        ('simple2', {
            'test_id': 'org.mozilla.unit-tests',
            'test_case_id': 'testcase125',
            'name': "Rendering test",
            'units': 'frames/s',
        }),
    ]

    def setUp(self):
        super(TestCaseConstructionTests, self).setUp()
        self.test = Test(test_id=self.test_id)
        self.test.save()

    def test_construction(self):
        test_case = TestCaseModel(
            test = self.test,
            test_case_id = self.test_case_id,
            name = self.name,
            units = self.units
        )
        test_case.save()
        self.assertEqual(self.name, test_case.name)
        self.assertEqual(self.test_case_id, test_case.test_case_id)
        self.assertEqual(self.name, test_case.name)
        self.assertEqual(self.units, test_case.units)

    def test_test_and_test_case_id_uniqueness(self):
        test_case = TestCaseModel(
            test = self.test,
            test_case_id = self.test_case_id)
        test_case.save()
        test_case2 = TestCaseModel(
            test = self.test,
            test_case_id = self.test_case_id)
        self.assertRaises(IntegrityError, test_case2.save)


class TestCaseUnicodeTests(TestCase):

    def test_test_case_with_id(self):
        """TestCase.test_case_id used when TestCase.name is not set"""
        obj = TestCaseModel(test_case_id="test123")
        self.assertEqual(unicode(obj), "test123")

    def test_test_case_with_name(self):
        """TestCase.name used when available"""
        obj = TestCaseModel(name="Test 123")
        self.assertEqual(unicode(obj), "Test 123")

    def test_test_case_with_id_and_name(self):
        """TestCase.name takes precedence over TestCase.test_case_id"""
        obj = TestCaseModel(name="Test 123", test_case_id="test123")
        self.assertEqual(unicode(obj), "Test 123")



# Copyright (C) 2010, 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Database models of the Dashboard application
"""

import datetime
import hashlib
import logging
import os
import simplejson
import traceback
import contextlib

from django.contrib.auth.models import User
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.db import models
from django.template import Template, Context
from django.utils.translation import ugettext as _
from django.utils.translation import ungettext

from django_restricted_resource.models  import RestrictedResource

from dashboard_app.helpers import BundleDeserializer
from dashboard_app.managers import BundleManager
from dashboard_app.repositories import RepositoryItem 
from dashboard_app.repositories.data_report import DataReportRepository
from dashboard_app.repositories.data_view import DataViewRepository

# Fix some django issues we ran into
from dashboard_app.patches import patch
patch()


def _help_max_length(max_length):
    return ungettext(
            u"Maximum length: {0} character",
            u"Maximum length: {0} characters",
            max_length).format(max_length)


class SoftwarePackage(models.Model):
    """
    Model for software packages.
    """
    name = models.CharField(
            max_length = 64,
            verbose_name = _(u"Package name"),
            help_text = _help_max_length(64))

    version = models.CharField(
            max_length = 64,
            verbose_name = _(u"Package version"),
            help_text = _help_max_length(32))

    class Meta:
        unique_together = (('name', 'version'))

    def __unicode__(self):
        return _(u"{name} {version}").format(
                name = self.name,
                version = self.version)

    @property
    def link_to_packages_ubuntu_com(self):
        return u"http://packages.ubuntu.com/{name}".format(name=self.name)


class SoftwarePackageScratch(models.Model):
    """
    Staging area for SoftwarePackage data.

    The code that keeps SoftwarePackage dumps data into here before more
    carefully inserting it into the real SoftwarePackage table.

    No data should ever be committed in this table.  It would be a temporary
    table, but oddities in how the sqlite DB-API wrapper handles transactions
    makes this impossible.
    """
    name = models.CharField(max_length=64)
    version = models.CharField(max_length=64)


class NamedAttribute(models.Model):
    """
    Model for adding generic named attributes
    to arbitrary other model instances.

    Example:
        class Foo(Model):
            attributes = generic.GenericRelation(NamedAttribute)
    """
    name = models.CharField(
            help_text = _help_max_length(32),
            max_length = 32)

    value = models.CharField(
            help_text = _help_max_length(512),
            max_length = 512)

    # Content type plumbing
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    def __unicode__(self):
        return _(u"{name}: {value}").format(
                name = self.name,
                value = self.value)

    class Meta:
        unique_together = (('object_id', 'name'))


class HardwareDevice(models.Model):
    """
    Model for hardware devices

    All devices are simplified into an instance of pre-defined class
    with arbitrary key-value attributes.
    """
    device_type = models.CharField(
            choices = (
                (u"device.cpu", _(u"CPU")),
                (u"device.mem", _(u"Memory")),
                (u"device.usb", _(u"USB device")),
                (u"device.pci", _(u"PCI device")),
                (u"device.board", _(u"Board/Motherboard"))),
            help_text = _(u"One of pre-defined device types"),
            max_length = 32,
            verbose_name = _(u"Device Type"),
            )

    description = models.CharField(
            help_text = _(u"Human readable device summary.") + " " + _help_max_length(256),
            max_length = 256,
            verbose_name = _(u"Description"),
            )

    attributes = generic.GenericRelation(NamedAttribute)

    def __unicode__(self):
        return self.description


class BundleStream(RestrictedResource):
    """
    Model for "streams" of bundles.

    Basically it's a named collection of bundles, like directory just
    without the nesting. A simple ACL scheme is also supported,
    a bundle may be uploaded by:
        - specific user when user field is set
        - users of a specific group when group field is set
        - anyone when neither user nor group is set
    """
    PATHNAME_ANONYMOUS = "anonymous"
    PATHNAME_PUBLIC = "public"
    PATHNAME_PRIVATE = "private"
    PATHNAME_PERSONAL = "personal"
    PATHNAME_TEAM = "team"

    slug = models.CharField(
            blank = True,
            help_text = (_(u"Name that you will use when uploading bundles.")
                + " " + _help_max_length(64)),
            max_length = 64,
            verbose_name = _(u"Slug"),
            )

    name = models.CharField(
            blank = True,
            help_text = _help_max_length(64),
            max_length = 64,
            verbose_name = _(u"Name"),
            )

    pathname = models.CharField(
            max_length = 128,
            editable = False,
            unique = True,
            )

    is_anonymous = models.BooleanField()

    def __unicode__(self):
        return self.pathname

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.bundle_list", [self.pathname])

    def get_test_run_count(self):
        return TestRun.objects.filter(bundle__bundle_stream=self).count()

    def clean(self):
        if self.is_anonymous and not self.is_public:
            raise ValidationError(
                'Anonymous streams must be public')
        return super(BundleStream, self).clean()

    def save(self, *args, **kwargs):
        """
        Save this instance.

        Calls self.clean() to ensure that constraints are met.
        Updates pathname to reflect user/group/slug changes.
        """
        self.pathname = self._calc_pathname()
        self.clean()
        return super(BundleStream, self).save(*args, **kwargs)

    def _calc_pathname(self):
        """
        Pseudo pathname-like ID of this stream.

        This pathname is user visible and will be presented to users
        when they want to interact with this bundle stream. The
        pathnames are unique and this is enforced at database level (the
        user and name are unique together).
        """
        if self.is_anonymous:
            parts = ['', self.PATHNAME_ANONYMOUS]
        else:
            if self.is_public:
                parts = ['', self.PATHNAME_PUBLIC]
            else:
                parts = ['', self.PATHNAME_PRIVATE]
            if self.user is not None:
                parts.append(self.PATHNAME_PERSONAL)
                parts.append(self.user.username)
            elif self.group is not None:
                parts.append(self.PATHNAME_TEAM)
                parts.append(self.group.name)
        if self.slug:
            parts.append(self.slug)
        parts.append('')
        return u"/".join(parts)

    @classmethod
    def parse_pathname(cls, pathname):
        """
        Parse BundleStream pathname.

        Returns user, group, slug, is_public, is_anonymous
        Raises ValueError if the pathname is not well formed
        """
        pathname_parts = pathname.split('/')
        if len(pathname_parts) < 3:
            raise ValueError("Pathname too short: %r" % pathname)
        if pathname_parts[0] != '':
            raise ValueError("Pathname must be absolute: %r" % pathname)
        if pathname_parts[1] == cls.PATHNAME_ANONYMOUS:
            user = None
            group = None
            slug = pathname_parts[2]
            correct_length = 2
            is_anonymous = True
            is_public = True
        else:
            is_anonymous = False
            if pathname_parts[1] == cls.PATHNAME_PUBLIC:
                is_public = True
            elif pathname_parts[1] == cls.PATHNAME_PRIVATE:
                is_public = False
            else:
                raise ValueError("Invalid pathname privacy designator:"
                        " %r (full pathname: %r)" % (pathname_parts[1],
                            pathname))
            if pathname_parts[2] == cls.PATHNAME_PERSONAL:
                if len(pathname_parts) < 4:
                    raise ValueError("Pathname too short: %r" % pathname)
                user = pathname_parts[3]
                group = None
                slug = pathname_parts[4]
                correct_length = 4
            elif pathname_parts[2] == cls.PATHNAME_TEAM:
                if len(pathname_parts) < 4:
                    raise ValueError("Pathname too short: %r" % pathname)
                user = None
                group = pathname_parts[3]
                slug = pathname_parts[4]
                correct_length = 4
            else:
                raise ValueError("Invalid pathname ownership designator:"
                        " %r (full pathname %r)" % (pathname[2],
                            pathname))
        if slug != '':
            correct_length += 1
        if pathname_parts[correct_length:] != ['']:
            raise ValueError("Junk after pathname: %r" % pathname)
        return user, group, slug, is_public, is_anonymous


class Bundle(models.Model):
    """
    Model for "Dashboard Bundles"
    """
    bundle_stream = models.ForeignKey(BundleStream,
            verbose_name = _(u"Stream"),
            related_name = 'bundles')

    uploaded_by = models.ForeignKey(User,
            verbose_name = _(u"Uploaded by"),
            help_text = _(u"The user who submitted this bundle"),
            related_name = 'uploaded_bundles',
            null = True,
            blank = True)

    uploaded_on = models.DateTimeField(
            verbose_name = _(u"Uploaded on"),
            editable = False,
            default = datetime.datetime.utcnow)

    is_deserialized = models.BooleanField(
            verbose_name = _(u"Is deserialized"),
            help_text = _(u"Set when document has been analyzed and loaded"
                " into the database"),
            editable = False)

    content = models.FileField(
            verbose_name = _(u"Content"),
            help_text = _(u"Document in Dashboard Bundle Format 1.0"),
            upload_to = 'bundles',
            null = True)

    content_sha1 = models.CharField(
            editable = False,
            max_length = 40,
            null = True,
            unique = True)

    content_filename = models.CharField(
            verbose_name = _(u"Content file name"),
            help_text = _(u"Name of the originally uploaded bundle"),
            max_length = 256)

    objects = BundleManager()

    def __unicode__(self):
        return _(u"Bundle {0}").format(self.content_sha1)

    class Meta:
        ordering = ['-uploaded_on']

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.bundle_detail", [self.bundle_stream.pathname, self.content_sha1])

    def get_permalink(self):
        return reverse("dashboard_app.views.redirect_to_bundle", args=[self.content_sha1])

    def save(self, *args, **kwargs):
        if self.content:
            try:
                self.content.open('rb')
                sha1 = hashlib.sha1()
                for chunk in self.content.chunks():
                    sha1.update(chunk)
                self.content_sha1 = sha1.hexdigest()
            finally:
                self.content.close()
        return super(Bundle, self).save(*args, **kwargs)

    def deserialize(self, prefer_evolution=False):
        """
        Deserialize the contents of this bundle.

        The actual implementation is _do_serialize() this function
        catches any exceptions it might throw and converts them to
        BundleDeserializationError instance. Any previous import errors are
        overwritten.

        Successful import also discards any previous import errors and
        sets is_deserialized to True.
        """
        if self.is_deserialized:
            return
        try:
            self._do_deserialize(prefer_evolution)
        except Exception as ex:
            import_error = BundleDeserializationError.objects.get_or_create(
                bundle=self)[0]
            import_error.error_message = str(ex)
            import_error.traceback = traceback.format_exc()
            import_error.save()
        else:
            if self.deserialization_error.count():
                self.deserialization_error.get().delete()
            self.is_deserialized = True
            self.save()

    def _do_deserialize(self, prefer_evolution):
        """
        Deserialize this bundle or raise an exception
        """
        helper = BundleDeserializer()
        helper.deserialize(self, prefer_evolution)

    def get_summary_results(self):
        if self.is_deserialized:
            stats = TestResult.objects.filter(
                test_run__bundle = self).values(
                    'result').annotate(
                        count=models.Count('result'))
            result = dict([
                (TestResult.RESULT_MAP[item['result']], item['count'])
                for item in stats])
            result['total'] = sum(result.values())
            return result

    def delete_files(self, save=False):
        """
        Delete all files related to this bundle.

        This is currently used in test code to clean up after testing.
        """
        self.content.delete(save=save)
        for test_run in self.test_runs.all():
            for attachment in test_run.attachments.all():
                attachment.content.delete(save=save)

    def get_sanitized_bundle(self):
        self.content.open()
        try:
            return SanitizedBundle(self.content)
        finally:
            self.content.close()


class SanitizedBundle(object):

    def __init__(self, stream):
        try:
            self.bundle_json = simplejson.load(stream)
            self.deserialization_error = None
        except simplejson.JSONDeserializationError as ex:
            self.bundle_json = None
            self.deserialization_error = ex
        self.did_remove_attachments = False
        self._sanitize()

    def get_human_readable_json(self):
        return simplejson.dumps(self.bundle_json, indent=4)

    def _sanitize(self):
        for test_run in self.bundle_json.get("test_runs", []):
            attachments = test_run.get("attachments")
            if isinstance(attachments, list):
                for attachment in attachments:
                    attachment["content"] = None
                    self.did_remove_attachments = True
            elif isinstance(attachments, dict):
                for name in attachments:
                    attachments[name] = None
                    self.did_remove_attachments = True


class BundleDeserializationError(models.Model):
    """
    Model for representing errors encountered during bundle
    deserialization. There is one instance per bundle limit due to
    unique = True. There used to be a OneToOne field but it didn't work
    with databrowse application.

    The relevant logic for managing this is in the Bundle.deserialize()
    """

    bundle = models.ForeignKey(
        Bundle,
        primary_key = True,
        unique = True,
        related_name = 'deserialization_error'
    )

    error_message = models.CharField(
        max_length = 1024
    )

    traceback = models.TextField(
        max_length = 1 << 15,
    )

    def __unicode__(self):
        return self.error_message


class Test(models.Model):
    """
    Model for representing tests.

    Test is a collection of individual test cases.
    """
    test_id = models.CharField(
        max_length = 64,
        verbose_name = _("Test ID"),
        unique = True,
        help_text = _help_max_length(64))

    name = models.CharField(
        blank = True,
        help_text = _help_max_length(64),
        max_length = 64,
        verbose_name = _(u"Name"))

    def __unicode__(self):
        return self.name or self.test_id

    @models.permalink
    def get_absolute_url(self):
        return ('dashboard_app.views.test_detail', [self.test_id])

    def count_results_without_test_case(self):
        return TestResult.objects.filter(
            test_run__test=self,
            test_case=None).count()

    def count_failures_without_test_case(self):
        return TestResult.objects.filter(
            test_run__test=self,
            test_case=None,
            result=TestResult.RESULT_FAIL).count()


class TestCase(models.Model):
    """
    Model for representing test cases.

    Test case is a unique component of a specific test.
    Test cases allow for relating to test results.
    """
    test = models.ForeignKey(
        Test,
        related_name='test_cases')

    test_case_id = models.CharField(
        help_text = _help_max_length(100),
        max_length = 100,
        verbose_name = _("Test case ID"))

    name = models.CharField(
        blank = True,
        help_text = _help_max_length(100),
        max_length = 100,
        verbose_name = _("Name"))

    units = models.CharField(
        blank = True,
        help_text = (_("""Units in which measurement value should be
                       interpreted in, for example <q>ms</q>, <q>MB/s</q> etc.
                       There is no semantical meaning inferred from the value of
                       this field, free form text is allowed. <br/>""")
                     + _help_max_length(100)),
        max_length = 100,
        verbose_name = _("Units"))

    class Meta:
        unique_together = (('test', 'test_case_id'))

    def __unicode__(self):
        return self.name or self.test_case_id

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.test_case.details", [self.test.test_id, self.test_case_id])

    def count_failures(self):
        return self.test_results.filter(result=TestResult.RESULT_FAIL).count()


class SoftwareSource(models.Model):
    """
    Model for representing source reference of a particular project
    """

    project_name = models.CharField(
        max_length = 32,
        help_text = _help_max_length(32),
        verbose_name = _(u"Project Name"),
    )
    branch_url = models.CharField(
        max_length = 256,
        help_text = _help_max_length(256),
        verbose_name = _(u"Branch URL"),
    )
    branch_vcs = models.CharField(
        max_length = 10,
        help_text = _help_max_length(10),
        verbose_name = _(u"Branch VCS"),
    )
    branch_revision = models.CharField(
        max_length = 128,
        help_text = _help_max_length(128),
        verbose_name = _(u"Branch Revision")
    )
    commit_timestamp = models.DateTimeField(
        blank=True,
        null=True,
        help_text = _(u"Date and time of the commit (optional)"),
        verbose_name = _(u"Commit Timestamp")
    )
    
    def __unicode__(self):
        return _(u"{project_name} from branch {branch_url} at revision {branch_revision}").format(
            project_name=self.project_name, branch_url=self.branch_url, branch_revision=self.branch_revision)

    @property
    def is_hosted_on_launchpad(self):
        return self.branch_url.startswith("lp:")

    @property
    def is_tag_revision(self):
        return self.branch_revision.startswith("tag:")

    @property
    def branch_tag(self):
        if self.is_tag_revision:
            return self.branch_revision[len("tag:"):]

    @property
    def link_to_project(self):
        return "http://launchpad.net/{project_name}".format(project_name=self.project_name)

    @property
    def link_to_branch(self):
        if self.is_hosted_on_launchpad:
            return "http://launchpad.net/{branch_url}/".format(branch_url=self.branch_url[len("lp:"):])

class TestRun(models.Model):
    """
    Model for representing test runs.

    Test run is an act of running a Test in a certain context. The
    context is described by the software and hardware environment.  In
    addition to those properties each test run can have arbitrary named
    properties for additional context that is not reflected in the
    database directly.

    Test runs have global identity exists beyond the lifetime of
    bundle that essentially encapsulates test run information should
    store the UUID that was generated at the time the document is made.
    the dashboard application. The software that prepares the dashboard
    """

    # Meta-data

    bundle = models.ForeignKey(
        Bundle,
        related_name = 'test_runs',
    )

    test = models.ForeignKey(
        Test,
        related_name = 'test_runs',
    )

    analyzer_assigned_uuid = models.CharField(
        help_text = _(u"You can use uuid.uuid1() to generate a value"),
        max_length = 36,
        unique = True,
        verbose_name = _(u"Analyzer assigned UUID"),
    )

    analyzer_assigned_date = models.DateTimeField(
        verbose_name = _(u"Analyzer assigned date"),
        help_text = _(u"Time stamp when the log was processed by the log"
                      " analyzer"),
    )

    import_assigned_date = models.DateTimeField(
        verbose_name = _(u"Import assigned date"),
        help_text = _(u"Time stamp when the bundle was imported"),
        auto_now_add = True,
    )

    time_check_performed = models.BooleanField(
        verbose_name = _(u"Time check performed"),
        help_text = _(u"Indicator on wether timestamps in the log file (and any "
                      "data derived from them) should be trusted.<br/>"
                      "Many pre-production or development devices do not "
                      "have a battery-powered RTC and it's not common for "
                      "development images not to synchronize time with "
                      "internet time servers.<br/>"
                      "This field allows us to track tests results that "
                      "<em>certainly</em> have correct time if we ever end up "
                      "with lots of tests results from 1972")
    )

    # Software Context

    sw_image_desc = models.CharField(
        blank = True,
        max_length = 100,
        verbose_name = _(u"Operating System Image"),
    )

    packages = models.ManyToManyField(
        SoftwarePackage,
        blank = True,
        related_name = 'test_runs',
        verbose_name = _(u"Software packages"),
    )

    sources = models.ManyToManyField(
        SoftwareSource,
        blank = True,
        related_name = 'test_runs',
        verbose_name = _(u"Software sources"),
    )

    # Hardware Context

    devices = models.ManyToManyField(
        HardwareDevice,
        blank = True,
        related_name = 'test_runs',
        verbose_name = _(u"Hardware devices"),
    )

    # Attributes

    attributes = generic.GenericRelation(NamedAttribute)

    # Attachments

    attachments = generic.GenericRelation('Attachment')

    def __unicode__(self):
        return _(u"Test run {0}").format(self.analyzer_assigned_uuid)

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.test_run_detail",
                [self.bundle.bundle_stream.pathname,
                 self.bundle.content_sha1,
                 self.analyzer_assigned_uuid])

    def get_permalink(self):
        return reverse("dashboard_app.views.redirect_to_test_run", args=[self.analyzer_assigned_uuid])

    def _get_summary_results(self, factor=3):
        stats = self.test_results.values('result').annotate(
            count=models.Count('result')).order_by()
        result = dict([
            (TestResult.RESULT_MAP[item['result']], item['count'])
            for item in stats])
        result['total'] = sum(result.values())
        result['total_multiplied'] = result['total'] * factor
        return result

    def get_summary_results(self):
        if not hasattr(self, '_cached_summary_results'):
            self._cached_summary_results = self._get_summary_results()
        return self._cached_summary_results

    class Meta:
        ordering = ['-import_assigned_date']


class Attachment(models.Model):
    """
    Model for adding attachments to any other models.
    """

    content = models.FileField(
        verbose_name = _(u"Content"),
        help_text = _(u"Attachment content"),
        upload_to = 'attachments',
        null = True)

    content_filename = models.CharField(
        verbose_name = _(u"Content file name"),
        help_text = _(u"Name of the original attachment"),
        max_length = 256)

    mime_type = models.CharField(
        verbose_name = _(u"MIME type"),
        max_length = 64)
    
    public_url = models.URLField(
        verbose_name = _(u"Public URL"),
        max_length = 512,
        blank = True)

    # Content type plumbing
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    def __unicode__(self):
        return self.content_filename

    def get_content_if_possible(self, mirror=False):
        if self.content:
            self.content.open()
            try:
                data = self.content.read()
            finally:
                self.content.close()
        elif self.public_url and mirror:
            import urllib
            stream = urllib.urlopen(self.public_url)
            try:
                data = stream.read()
            except:
                data = None
            else:
                from django.core.files.base import ContentFile
                self.content.save(
                    "attachment-{0}.txt".format(self.pk),
                    ContentFile(data))
            finally:
                stream.close()
        else:
            data = None
        return data

    def is_test_run_attachment(self):
        if (self.content_type.app_label == 'dashboard_app' and
            self.content_type.model == 'testrun'):
            return True

    @property
    def test_run(self):
        if self.is_test_run_attachment():
            return self.content_object

    @models.permalink
    def get_absolute_url(self):
        if self.is_test_run_attachment():
            return ("dashboard_app.views.attachment_detail",
                    [self.test_run.bundle.bundle_stream.pathname,
                     self.test_run.bundle.content_sha1,
                     self.test_run.analyzer_assigned_uuid,
                     self.pk])


class TestResult(models.Model):
    """
    Model for representing test results.
    """

    RESULT_PASS = 0
    RESULT_FAIL = 1
    RESULT_SKIP = 2
    RESULT_UNKNOWN = 3

    RESULT_MAP = {
        RESULT_PASS: 'pass',
        RESULT_FAIL: 'fail',
        RESULT_SKIP: 'skip',
        RESULT_UNKNOWN: 'unknown'
    }

    # Context information

    test_run = models.ForeignKey(
        TestRun,
        related_name = "test_results"
    )

    test_case = models.ForeignKey(
        TestCase,
        related_name = "test_results",
        null = True,
        blank = True
    )

    @property
    def test(self):
        return self.test_run.test

    # Core attributes

    result = models.PositiveSmallIntegerField(
        verbose_name = _(u"Result"),
        help_text = _(u"Result classification to pass/fail group"),
        choices = (
            (RESULT_PASS, _(u"Test passed")),
            (RESULT_FAIL, _(u"Test failed")),
            (RESULT_SKIP, _(u"Test skipped")),
            (RESULT_UNKNOWN, _(u"Unknown outcome")))
    )

    measurement = models.DecimalField(
        blank = True,
        decimal_places = 10,
        help_text = _(u"Arbitrary value that was measured as a part of this test."),
        max_digits = 20,
        null = True,
        verbose_name = _(u"Measurement"),
    )

    # Misc attributes

    filename = models.CharField(
        blank = True,
        max_length = 1024,
        null = True,
    )

    lineno = models.PositiveIntegerField(
        blank = True,
        null = True
    )

    message = models.TextField(
        blank = True,
        max_length = 1024,
        null = True
    )

    microseconds = models.BigIntegerField(
        blank = True,
        null = True
    )

    timestamp = models.DateTimeField(
        blank = True,
        null = True
    )

    relative_index = models.PositiveIntegerField(
        help_text = _(u"The relative order of test results in one test run")
    )

    def __unicode__(self):
        return "Result {0}/{1}".format(self.test_run.analyzer_assigned_uuid, self.relative_index)

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.test_result_detail", [
            self.test_run.bundle.bundle_stream.pathname,
            self.test_run.bundle.content_sha1,
            self.test_run.analyzer_assigned_uuid,
            self.relative_index,
        ])

    def get_permalink(self):
        return reverse("dashboard_app.views.redirect_to_test_result",
                       args=[self.test_run.analyzer_assigned_uuid,
                             self.relative_index])

    @property
    def result_code(self):
        """
        Stable textual result code that does not depend on locale
        """
        return self.RESULT_MAP[self.result]

    # units (via test case)

    @property
    def units(self):
        try:
            return self.test_case.units
        except TestCase.DoesNotExist:
            return None

    # Attributes

    attributes = generic.GenericRelation(NamedAttribute)

    # Duration property

    def _get_duration(self):
        if self.microseconds is None:
            return None
        else:
            return datetime.timedelta(microseconds = self.microseconds)

    def _set_duration(self, duration):
        if duration is None:
            self.microseconds = None
        else:
            if not isinstance(duration, datetime.timedelta):
                raise TypeError("duration must be a datetime.timedelta() instance")
            self.microseconds = (
                duration.microseconds +
                (duration.seconds * 10 ** 6) +
                (duration.days * 24 * 60 * 60 * 10 ** 6))

    duration = property(_get_duration, _set_duration)

    def related_attachment_available(self):
        """
        Check if there is a log file attached to the test run that has
        the same filename as log filename recorded in the result here.
        """
        try:
            self.related_attachment()
            return True
        except Attachment.DoesNotExist:
            return False

    def related_attachment(self):
        return self.test_run.attachments.get(content_filename=self.filename)

    class Meta:
        ordering = ['relative_index']
        order_with_respect_to = 'test_run'


class DataView(RepositoryItem):
    """
    Data view, a container for SQL query and optional arguments
    """

    repository = DataViewRepository()
    
    def __init__(self, name, backend_queries, arguments, documentation, summary):
        self.name = name
        self.backend_queries = backend_queries
        self.arguments = arguments
        self.documentation = documentation
        self.summary = summary

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return "<DataView name=%r>" % (self.name,)

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.data_view_detail", [self.name])

    def _get_connection_backend_name(self, connection):
        backend = str(type(connection))
        if "sqlite" in backend:
            return "sqlite"
        elif "postgresql" in backend:
            return "postgresql"
        else:
            return ""

    def get_backend_specific_query(self, connection):
        """
        Return BackendSpecificQuery for the specified connection
        """
        sql_backend_name = self._get_connection_backend_name(connection)
        try:
            return self.backend_queries[sql_backend_name]
        except KeyError:
            return self.backend_queries.get(None, None)

    def lookup_argument(self, name):
        """
        Return Argument with the specified name

        Raises LookupError if the argument cannot be found
        """
        for argument in self.arguments:
            if argument.name == name:
                return argument
        raise LookupError(name)

    @classmethod
    def get_connection(cls):
        """
        Get the appropriate connection for data views
        """
        from django.db import connection, connections
        from django.db.utils import ConnectionDoesNotExist
        try:
            return connections['dataview']
        except ConnectionDoesNotExist:
            logging.warning("dataview-specific database connection not available, dataview query is NOT sandboxed")
            return connection  # NOTE: it's connection not connectionS (the default connection)

    def __call__(self, connection, **arguments):
        # Check if arguments have any bogus names
        valid_arg_names = frozenset([argument.name for argument in self.arguments])
        for arg_name in arguments:
            if arg_name not in valid_arg_names:
                raise TypeError("Data view %s has no argument %r" % (self.name, arg_name))
        # Get the SQL template for our database connection
        query = self.get_backend_specific_query(connection)
        if query is None:
            raise LookupError("Specified data view has no SQL implementation "
                              "for current database")
        # Replace SQL aruments with django placeholders (connection agnostic)
        template = query.sql_template
        template = template.replace("%", "%%")
        # template = template.replace("{", "{{").replace("}", "}}")
        sql = template.format(
            **dict([
                (arg_name, "%s")
                for arg_name in query.argument_list]))
        # Construct argument list using defaults for missing values
        sql_args = [
            arguments.get(arg_name, self.lookup_argument(arg_name).default)
            for arg_name in query.argument_list]
        with contextlib.closing(connection.cursor()) as cursor:
            # Execute the query with the specified arguments
            cursor.execute(sql, sql_args)
            # Get and return the results
            rows = cursor.fetchall()
            columns = cursor.description
            return rows, columns


class DataReport(RepositoryItem):
    """
    Data reports are small snippets of xml that define
    a limited django template.
    """
    
    repository = DataReportRepository()

    def __init__(self, **kwargs):
        self._html = None
        self._data = kwargs

    def __unicode__(self):
        return self.title

    def __repr__(self):
        return "<DataReport name=%r>" % (self.name,)

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.report_detail", [self.name])

    def _get_raw_html(self):
        pathname = os.path.join(self.base_path, self.path)
        try:
            with open(pathname) as stream:
                return stream.read()
        except (IOError, OSError) as ex:
            logging.error("Unable to load DataReport HTML file from %r: %s", pathname, ex)
            return ""

    def _get_html_template(self):
        return Template(self._get_raw_html())

    def _get_html_template_context(self):
        from django.conf import settings
        return Context({
            "API_URL": reverse("dashboard_app.views.dashboard_xml_rpc_handler"),
            "STATIC_URL": settings.STATIC_URL
        })

    def get_html(self):
        from django.conf import settings
        DEBUG = getattr(settings, "DEBUG", False)
        if self._html is None or DEBUG is True:
            template = self._get_html_template()
            context = self._get_html_template_context()
            self._html = template.render(context)
        return self._html

    @property
    def title(self):
        return self._data['title']

    @property
    def path(self):
        return self._data['path']

    @property
    def name(self):
        return self._data['name']

    @property
    def bug_report_url(self):
        return self._data.get('bug_report_url')

    @property
    def author(self):
        return self._data.get('author')


class ImageHealth(object):

    def __init__(self, rootfs_type, hwpack_type):
        self.rootfs_type = rootfs_type
        self.hwpack_type = hwpack_type

    @models.permalink
    def get_absolute_url(self):
        return ("dashboard_app.views.image_status_detail", [
            self.rootfs_type, self.hwpack_type])

    def get_tests(self):
        return Test.objects.filter(test_runs__in=self.get_test_runs()).distinct()

    def current_health_for_test(self, test):
        test_run = self.get_current_test_run(test)
        test_results = test_run.test_results
        fail_count = test_results.filter(
            result=TestResult.RESULT_FAIL).count()
        pass_count = test_results.filter(
            result=TestResult.RESULT_PASS).count()
        total_count = test_results.count()
        return {
            "test_run": test_run,
            "total_count": total_count,
            "fail_count": fail_count,
            "pass_count": pass_count,
            "other_count": total_count - (fail_count + pass_count),
            "fail_percent": fail_count * 100.0 / total_count if total_count > 0 else None,
        }
    
    def overall_health_for_test(self, test):
        test_run_list = self.get_all_test_runs_for_test(test)
        test_result_list = TestResult.objects.filter(test_run__in=test_run_list)
        fail_result_list = test_result_list.filter(result=TestResult.RESULT_FAIL)
        pass_result_list = test_result_list.filter(result=TestResult.RESULT_PASS)

        total_count = test_result_list.count()
        fail_count = fail_result_list.count()
        pass_count = pass_result_list.count()
        fail_percent = fail_count * 100.0 / total_count if total_count > 0 else None
        return {
            "total_count": total_count,
            "total_run_count": test_run_list.count(),
            "fail_count": fail_count,
            "pass_count": pass_count,
            "other_count": total_count - fail_count - pass_count,
            "fail_percent": fail_percent,
        }

    def get_recent_test_runs(self, last_n_days=7 * 2):
        until = datetime.datetime.now()
        since = until - datetime.timedelta(days=last_n_days)
        return self.get_test_runs(
        ).select_related(
        ).filter(
            analyzer_assigned_date__range=(since, until)
        ).order_by('-analyzer_assigned_date')

    def get_chart_data(self):
        return TestResult.objects.select_related(
        ).filter(
            test_run__in=[run.pk for run in self.get_recent_test_runs().only('id').order_by()]
        ).values(
            'test_run__analyzer_assigned_date'
        ).extra(
            select={'bucket': 'date(analyzer_assigned_date)'},
        ).values(
            'bucket', 'result'
        ).annotate(
            count=models.Count('result')
        ).order_by('result')

    def get_test_runs(self):
        return TestRun.objects.select_related(
        ).filter(
            bundle__bundle_stream__pathname="/anonymous/lava-daily/"
        ).filter(
            attributes__name='rootfs.type'
        ).filter(
            attributes__value=self.rootfs_type
        ).filter(
            attributes__name='hwpack.type'
        ).filter(
            attributes__value=self.hwpack_type
        )

    def get_current_test_run(self, test):
        return self.get_all_test_runs_for_test(test).order_by('-analyzer_assigned_date')[0]

    def get_all_test_runs_for_test(self, test):
        return self.get_test_runs().filter(test=test)

    @classmethod
    def get_rootfs_list(self):
        rootfs_list = [
            attr['value'] 
            for attr in NamedAttribute.objects.filter(
                name='rootfs.type').values('value').distinct()]
        try:
            rootfs_list.remove('android')
        except ValueError:
            pass
        return rootfs_list

    @classmethod
    def get_hwpack_list(self):
        hwpack_list = [
            attr['value'] 
            for attr in NamedAttribute.objects.filter(
                name='hwpack.type').values('value').distinct()]
        return hwpack_list



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

from django.core.urlresolvers import reverse
import logging


class BreadCrumb(object):

    def __init__(self, name, parent=None, needs=None):
        self.name = name
        self.view = None
        self.parent = parent
        self.needs = needs or []

    def __repr__(self):
        return "<BreadCrumb name=%r view=%r parent=%r>" % (
            self.name, self.view, self.parent)

    def __call__(self, view):
        self.view = view
        view._bread_crumb = self
        return view

    def get_name(self, kwargs):
        try:
            return self.name.format(**kwargs)
        except:
            logging.exception("Unable to construct breadcrumb name for view %r", self.view)
            raise

    def get_absolute_url(self, kwargs):
        try:
            return reverse(self.view, args=[kwargs[name] for name in self.needs])
        except:
            logging.exception("Unable to construct breadcrumb URL for view %r", self.view)
            raise


class LiveBreadCrumb(object):

    def __init__(self, bread_crumb, kwargs):
        self.bread_crumb = bread_crumb
        self.kwargs = kwargs

    def get_name(self):
        return self.bread_crumb.get_name(self.kwargs)

    def get_absolute_url(self):
        return self.bread_crumb.get_absolute_url(self.kwargs)


class BreadCrumbTrail(object):

    def __init__(self, bread_crumb_list, kwargs):
        self.bread_crumb_list = bread_crumb_list
        self.kwargs = kwargs

    def __iter__(self):
        for bread_crumb in self.bread_crumb_list:
            yield LiveBreadCrumb(bread_crumb, self.kwargs)

    @classmethod
    def leading_to(cls, view, **kwargs):
        lst = []
        while view is not None:
            lst.append(view._bread_crumb)
            view = view._bread_crumb.parent
        lst.reverse()
        return cls(lst, kwargs or {})




#!/usr/bin/python
# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Small helper module that aids in writing unit tests. It allows to
construct objects or call functions without any arguments by looking up
the required arguments from a helper object or from the class itself.
"""

from contextlib import contextmanager


class CallHelper(object):
    """
    Helper for making easy-to-call functions that would otherwise
    require multiple (possibly complex) arguments to call correctly.

    The helper calls the provided function and takes care of any non-default
    arguments that were not explicitly provided by getting 'good' values from a
    helper object.

    Example:
    >>> def add(a, b): return a + b
    >>> class dummy_values:
    ...     a = 1
    ...     b = 2
    >>> add = CallHelper(add, dummy_values())

    In reality the code below calls add(dummy.a, dummy.b) dummy is the
    instance of dummy_values.
    >>> add()
    3

    You can still call the function as it was the original
    >>> add(5, -1)
    4

    Let's try something more complicated.
    >>> def inc(value, increment=1):
    ...     return value + increment
    >>> class inc_dummy(object):
    ...     value = 5
    ...     increment = 2

    Here the function already has a default value for one of the
    arguments. This argument will not be obtained from the helper
    object.
    >>> inc = CallHelper(inc, inc_dummy())
    >>> inc()
    6

    This is same as using DEFAULT_VALUE
    >>> inc(increment=inc.DEFAULT_VALUE)
    6

    If you want to override this behaviour and force dummy values for
    default argument you can use special DUMMY_VALUE value.
    >>> inc(increment=inc.DUMMY_VALUE)
    7

    There are two special values you can use DEFAULT_VALUE and DUMMY_VALUE. Passing
    DEFAULT_VALUE simply instructs the helper not to provide a value from the
    defaults even if we can. Passing DUMMY_VALUE does the opposite. Having
    both makes sense when you consider positional arguments.
    >>> def add3(a, b, c):
    ...     return a + b + c
    >>> class add3_dummy(object):
    ...     a = 1
    ...     b = 2
    ...     c = 3
    >>> add3 = CallHelper(add3, add3_dummy)
    >>> add3()
    6

    What if I want to specify the middle argument while the rest gets
    filled with dummies? Simple, use DUMMY_VALUE positional arguments.
    >>> add3(add3.DUMMY_VALUE, -2, add3.DUMMY_VALUE)
    2

    Another way of doing that is to use keyword arguments
    >>> add3(b=-2)
    2
    """
    DEFAULT_VALUE = object()
    DUMMY_VALUE = object()

    def __init__(self, func, dummy, dummy_preference=False):
        """
        Initialize call helper to wrap function `func' and supply values
        from properties of `dummy' object.

        Dummy preference defaults to False. That is, by default any
        argument that has a default value will prefer the default value
        rather than the dummy value obtained from dummy object.
        """
        if func.func_code.co_flags & 0x4:
            raise ValueError("Functions with variable argument lists "
                    "are not supported")
        if func.func_code.co_flags & 0x8:
            raise ValueError("Functions with variable keyword arguments "
                    "are not supported")
        self._func = func
        self._dummy = dummy
        self._dummy_preference = dummy_preference
        self._args = func.func_code.co_varnames[:func.func_code.co_argcount]
        self._args_with_defaults = dict(
                zip(self._args[-len(func.func_defaults):] if
                    func.func_defaults else (), func.func_defaults or ()))

    def _get_dummy_for(self, arg_name):
        """
        Get dummy value for given argument.

        Attributes are extracted from arbitrary objects
        >>> class dummy:
        ...     a = 1
        >>> CallHelper(lambda x: x, dummy)._get_dummy_for('a')
        1

        Or from dictionary items (for one-liner)
        >>> CallHelper(lambda x: x, {'a': 1})._get_dummy_for('a')
        1
        """
        try:
            return getattr(self._dummy, arg_name)
        except AttributeError:
            try:
                return self._dummy[arg_name]
            except KeyError:
                raise ValueError("Dummy %s does not have dummy value for %s" % (
                    self._dummy, arg_name))

    def _fill_args(self, *args, **kwargs):
        a_out = []
        used_kwargs = set()
        # Walk through all arguments of the original function. Ff the
        # argument is present in `args' then use it. If we run out of
        # positional arguments look for keyword arguments with the same
        # name.
        for i, arg_name in enumerate(self._args):
            # find the argument
            if i < len(args):
                # positional arguments get passed as-is
                arg = args[i]
            elif arg_name in kwargs:
                # keyword arguments take over
                arg = kwargs[arg_name]
                # also remember we got it from a keyword argument
                used_kwargs.add(arg_name)
            else:
                # otherwise the function defaults kick in
                # with a yet another fall-back to special DUMMY_VALUE value

                # Note that there is a special configuration for
                # preferring default values over dummy values.
                if self._dummy_preference:
                    arg = self.DUMMY_VALUE
                else:
                    arg = self._args_with_defaults.get(arg_name, self.DUMMY_VALUE)
            # resolve the argument
            if arg is self.DEFAULT_VALUE:
                if arg_name not in self._args_with_defaults:
                    raise ValueError("You passed DEFAULT_VALUE argument to %s "
                            "which has no default value" % (arg_name,))
                arg = self._args_with_defaults[arg_name]
            elif arg is self.DUMMY_VALUE:
                arg = self._get_dummy_for(arg_name)
            # store the argument
            a_out.append(arg)
        # Now check if we have too many / not enough arguments
        if len(a_out) != len(self._args):
            raise TypeError("%s takes exactly %d argument, %d given" % (
                self._func, len(self._args), len(args)))
        # Now check keyword arguments
        for arg_name in kwargs:
            # Check for duplicate definitions of positional/keyword
            # arguments
            if arg_name in self._args and arg_name not in used_kwargs:
                raise TypeError("%s() got multiple values for keyword "
                        "argument '%s'" % (self._func.func_name, arg_name))

            # Look for stray keyword arguments
            if arg_name not in self._args:
                raise TypeError("%s() got an unexpected keyword "
                        "argument '%s'" % (self._func.func_name, arg_name))
        # We're done
        return a_out

    def __call__(self, *args, **kwargs):
        """
        Call the original function passing dummy values for all
        arguments """
        a_out = self._fill_args(*args, **kwargs)
        # We're done, let's call the function now! Note that we don't
        # use kw_args as we resolve keyword arguments ourselves, this
        # would be only useful for functions that support *args and
        # **kwargs style variable arguments which CallHelper does not
        # yet support.
        return self._func(*a_out)

    @contextmanager
    def dummy_preference(self, dummy_preference=True):
        """
        Context manager that allows dummy values to override (be
        preferred to) default values for unspecified arguments.

        Example:
        >>> def inc(a, b=1): return a + b
        >>> inc = CallHelper(inc, {'b': -1})
        >>> inc(5)
        6

        All code executed under this context manager will prefer dummy
        values (specified as the dummy argument to the CallHelper
        constructor) even if a default argument value is available.
        >>> with inc.dummy_preference():
        ...     inc(5)
        4

        This behaviour automatically reverts after the 'with' block
        >>> inc(5)
        6
        """
        old_preference = self._dummy_preference
        try:
            self._dummy_preference = dummy_preference
            yield
        finally:
            self._dummy_preference = old_preference



class ObjectFactory(CallHelper):
    """
    Helper class for making objects

    This class allows to easily construct a dummy instances by building
    a call_helper which will call the constructor (a rather tricky thing
    to do).

    All non-default values for the constructor will be fetched from a
    helper object. By default you can use a convention where a nested
    class called '_Dummy' contains all the properties you'd need to make
    instances.
    >>> class Person(object):
    ...     class _Dummy:
    ...         name = "Joe"
    ...     def __init__(self, name):
    ...         self.name = name
    >>> factory = ObjectFactory(Person)
    >>> person = factory()
    >>> person.name
    'Joe'

    As with CallHelper you can override dummy values by passing any
    positional or keyword arguments.
    >>> person = factory(name="Bob")
    >>> person.name
    'Bob'

    If you want you can store dummy values in separate class and pass it
    along to the ObjectFactory constructor.
    >>> class MyDummy:
    ...     name = "Alice"
    >>> factory = ObjectFactory(Person, MyDummy)
    >>> person = factory()
    >>> person.name
    'Alice'

    In addition the factory expoes the dummy values it is using via the
    `dummy' property. This can be useful in unit tests where you don't
    want to worry about the particular value but need to check if it can
    be stored, retrieved, etc.
    >>> name = factory.dummy.name
    >>> name
    'Alice'
    """
    def __init__(self, cls, dummy_cls=None):
        """
        Initialize ObjectFactory to create instances of class `cls'.
        If specified, dummy values for required arguments will come from
        instances of `dummy_cls'. Otherwise the class must have a
        '_Dummy' nested class that will be used instead.
        """
        if dummy_cls is None:
            if not hasattr(cls, '_Dummy'):
                raise ValueError("Class %s needs to have a nested class"
                        " called _Dummy" % (cls,))
            dummy_cls = cls._Dummy
        self._cls = cls
        self._dummy = dummy_cls()
        super(ObjectFactory, self).__init__(cls.__init__, self._dummy)
        # remove 'self' from argument list
        self._args = self._args[1:]

    @property
    def dummy(self):
        """ Helper property for read-only access to dummy values """
        return self._dummy

    def __call__(self, *args, **kwargs):
        """ Construct new object instance """
        a_out = self._fill_args(*args, **kwargs)
        return self._cls(*a_out)

    def full_construct(self, *args, **kwargs):
        """ Construct new object instance """
        with self.dummy_preference():
            a_out = self._fill_args(*args, **kwargs)
        return self._cls(*a_out)

class DummyValues(object):
    """
    Class for holding values used to initialize instances created with
    ObjectFactoryMixIn
    """
    def __init__(self, dummy_cls):
        self._ctor_args = {}
        # Instantiate the class to support properties
        dummy_obj = dummy_cls()
        for attr_name in dir(dummy_obj):
            if not attr_name.startswith("_"):
                self._ctor_args[attr_name] = getattr(dummy_obj, attr_name)

    def __getattr__(self, name):
        return self._ctor_args[name]

    def get_ctor_args(self):
        return self._ctor_args


class ObjectFactoryMixIn(object):
    """
    Helper mix-in for various unittest.TestCase like classes.  Allows for
    convenient encapsulation and specification of dummy values for
    constructing objects of any class without cluttering the code of the
    test cases.

    Example:
    >>> class Person(object):
    ...     def __init__(self, name):
    ...         self.name = name

    >>> class Test(ObjectFactoryMixIn):
    ...     class Dummy:
    ...         class Person:
    ...             name = "Joe"

    Simplest form is just to call the make() method with a class All the
    constructor arguments will be then taken from Test.Dummy.Person
    object (whatever it is, it's a class in this example)
    >>> person = Test().make(Person)
    >>> person.name
    'Joe'

    If you want to ensure that the values are stored or used by the
    constructor properly but you don't want to hard-code 'good' values
    in your code you can look at the dummy object that was used to
    provide the values.
    >>> dummy, person = Test().make_and_get_dummy(Person)
    >>> dummy.name
    'Joe'

    So essentially the (test case) code will do something similar to
    this:
    >>> person.name == dummy.name
    True
    """

    def make(self, cls, dummy_cls = None):
        """
        Make an object using make_and_get_dummy() and discard the dummy.
        """
        dummy, obj = self.make_and_get_dummy(cls, dummy_cls)
        return obj

    def make_and_get_dummy(self, cls, dummy_cls=None):
        """
        Make an object using the specified dummy_cls or find the
        container of default values in the encompassing class.

        class MyTest(..., ObjectFactoryMixIn):
            class Dummy:
                class SomeClassYouWantToCreate:
                    ctr_arg1 = 1
        """
        if dummy_cls is None:
            if hasattr(self, 'Dummy') and hasattr(self.Dummy, cls.__name__):
                dummy_cls = getattr(self.Dummy, cls.__name__)
            else:
                raise ValueError("%r does not have nested class Dummy.%s" % (
                    self.__class__, cls.__name__))
        dummy = DummyValues(dummy_cls)
        obj = cls(**dummy.get_ctor_args())
        return dummy, obj


if __name__ == '__main__':
    import doctest
    doctest.testmod()





# Copyright (C) 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.


from xml.sax import parseString

from dashboard_app.repositories import Repository, Undefined, Object
from dashboard_app.repositories.common import BaseContentHandler


class _DataReportHandler(BaseContentHandler):
    """
    ContentHandler subclass for parsing DataReport documents
    """
    
    def startDocument(self):
        # Classic-classes 
        BaseContentHandler.startDocument(self)
        # Data report object
        self.obj = Object()

    def endDocument(self):
        if self.obj.name is Undefined:
            raise ValueError("No data view definition found")
        if self.obj.title is Undefined:
            raise ValueError("Data report without a title")
        if self.obj.path is Undefined:
            raise ValueError("Data report without a path")

    def startElement(self, name, attrs):
        if name == "data-report":
            if "name" not in attrs:
                raise ValueError("Data report without a name")
            self.obj.name = attrs["name"]
        elif name == "title" or name == "path":
            self._start_text()
                
    def endElement(self, name):
        if name == "title":
            self.obj.title = self._end_text()
        elif name == "path":
            self.obj.path = self._end_text()


class DataReportRepository(Repository):

    @property
    def settings_variable(self):
        return "DATAREPORT_DIRS"

    def load_from_xml_string(self, text):
        handler = _DataReportHandler()
        parseString(text, handler)
        return self.item_cls(
            name=handler.obj.name,
            title=handler.obj.title,
            path=handler.obj.path)


__all__ = [
    "DataReportRepository",
]



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for the TestResult model
"""
import datetime

from django_testscenarios.ubertest import TestCase, TestCaseWithScenarios

from dashboard_app.models import TestRun, TestResult


class TestResultDurationTests(TestCaseWithScenarios):

    scenarios = [
        ('none_is_null', {
            'duration': None,
            'microseconds': None,
        }),
        ('0_is_0', {
            'duration': datetime.timedelta(days=0, seconds=0, microseconds=0),
            'microseconds': 0,
        }),
        ('microseconds_are_just_microseconds', {
            'duration': datetime.timedelta(microseconds=1),
            'microseconds': 1,
        }),
        ('second_is_10e6_microseconds', {
            'duration': datetime.timedelta(seconds=1),
            'microseconds': 10**6,
        }),
        ('day_is_24_times_60_times_60_times_10e6_microseconds', {
            'duration': datetime.timedelta(days=1),
            'microseconds': 24 * 60 * 60 * 10 ** 6,
        }),
        ('microseconds_seconds_and_days_are_used', {
            'duration': datetime.timedelta(days=1, seconds=1, microseconds=1),
            'microseconds': (
                24 * 60 * 60 * (10 ** 6) +
                10 ** 6 +
                1)
        }),
    ]

    def test_duration_to_microseconds(self):
        obj = TestResult()
        obj.duration = self.duration
        self.assertEqual(self.microseconds, obj.microseconds)

    def test_microseconds_to_duration(self):
        obj = TestResult()
        obj.microseconds = self.microseconds
        self.assertEqual(self.duration, obj.duration)


class TestResultUnicodeTests(TestCase):

    def test_test_result__pass(self):
        test_run = TestRun(analyzer_assigned_uuid="00000000-0000-0000-0000-000000000004")
        test_result = TestResult(test_run=test_run, relative_index=1)
        self.assertIn("00000000-0000-0000-0000-000000000004/1", unicode(test_result))



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Helper functions for making fixtures that setup specific environment
"""

from contextlib import contextmanager

from django.contrib.auth.models import (User, Group)

from dashboard_app.models import (Bundle, BundleStream)


class test_loop(object):
    """
    Support class that tells you something about a test crashing when
    the actual test values depend on a loop value
    """

    def __init__(self, source):
        self._iter = iter(source)
        self._last = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            import logging
            logging.exception("Exception in test_loop on iteration: %r", self._last)

    def __iter__(self):
        return self

    def next(self):
        self._last = next(self._iter)
        return self._last


def create_bundle_stream(pathname):
    """
    Create, or get an existing bundle stream designated by the provided
    pathname. The pathname is parsed and decomposed to determine the
    user/group and slug. Users and groups are created if necessary.
    """
    try:
        return BundleStream.objects.get(pathname=pathname)
    except BundleStream.DoesNotExist:
        user_username, group_name, slug, is_public, is_anonymous = BundleStream.parse_pathname(pathname)
        if user_username is None and group_name is None:
            # Here we trick a little - since the introduction of the
            # django-restricted-resource each object has a principal
            # owner - either a user or a group. This information can be
            # conveyed from the pathname _except_ for anonymous streams
            # that are a remnant of the past. For those streams we just
            # create a dummy user.
            user_username = "anonymous-stream-owner"
        if user_username is not None:
            user = User.objects.get_or_create(username=user_username)[0]
        else:
            user = None
        if group_name is not None:
            group = Group.objects.get_or_create(name=group_name)[0]
        else:
            group = None
        bundle_stream = BundleStream.objects.create(
            user=user, group=group, slug=slug,
            is_public=is_public, is_anonymous=is_anonymous)
        bundle_stream.save()
        return bundle_stream


def create_bundle(pathname, content, content_filename):
    """"
    Create bundle with the specified content and content_filename and
    place it in a bundle stream designated by the specified pathname.
    Bundle stream is created if required.
    """
    return Bundle.objects.create_with_content(
        bundle_stream=create_bundle_stream(pathname),
        uploaded_by=None,
        content_filename=content_filename,
        content=content)


@contextmanager
def created_bundle_streams(pathnames):
    """
    Helper context manager that creates bundle streams according to
    specification.

    `pathnames`: list of pathnames to create
        List of values to create_bundle_stream()

    yields: list of BundleStream
        List of created bundle stream objects
    """
    bundle_streams = []
    for pathname in pathnames:
        bundle_streams.append(create_bundle_stream(pathname))
    yield bundle_streams


@contextmanager
def created_bundles(spec):
    """
    Helper context manager that creates bundles according to specification

    spec is a list of 3-element tuples:
        pathname: string, bundle stream pathname (all variants supported)
        content: string, text of the bundle
        content_filename: string

    yields: list of created bundles
    """
    bundles = []
    for pathname, content_filename, content in spec:
        bundles.append(
            create_bundle(pathname, content, content_filename))
    yield bundles
    for bundle in bundles:
        bundle.delete_files()



from django import template


register = template.Library()


class CallNode(template.Node):
    def __init__(self, func, args, name, nodelist):
        self.func = func
        self.args = args
        self.name = name
        self.nodelist = nodelist

    def __repr__(self):
        return "<CallNode>"

    def _lookup_func(self, context):
        parts = self.func.split('.')
        current = context[parts[0]] 
        for part in parts[1:]:
            if part.startswith("_"):
                raise ValueError(
                    "Function cannot traverse private implementation attributes")
            current = getattr(current, part)
        return current

    def render(self, context):
        try:
            func = self._lookup_func(context) 
            values = [template.Variable(arg).resolve(context) for arg in self.args]
            context.push()
            context[self.name] = func(*values)
            output = self.nodelist.render(context)
            context.pop()
            return output
        except Exception as ex:
            import logging
            logging.exception("Unable to call %s with %r: %s",
                              self.func, self.args, ex)
            raise

def do_call(parser, token):
    """
    Adds a value to the context (inside of this block) for caching and easy
    access.

    For example::

        {% call func 1 2 3 as result %}
            {{ result }}
        {% endcall %}
    """
    bits = list(token.split_contents())
    if len(bits) < 2 or bits[-2] != "as":
        raise template.TemplateSyntaxError(
            "%r expected format is 'call func [args] [as name]'" % bits[0])
    func = bits[1]
    args = bits[2:-2]
    name = bits[-1]
    nodelist = parser.parse(('endcall',))
    parser.delete_first_token()
    return CallNode(func, args, name, nodelist)

do_call = register.tag('call', do_call)



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Views for the Dashboard application
"""

import json

from django.db.models.manager import Manager
from django.db.models.query import QuerySet
from django.http import Http404, HttpResponse
from django.shortcuts import render_to_response, redirect, get_object_or_404
from django.template import RequestContext
from django.views.generic.list_detail import object_list, object_detail

from dashboard_app.models import (
    Attachment,
    Bundle,
    BundleStream,
    DataReport,
    DataView,
    ImageHealth,
    Test,
    TestResult,
    TestRun,
)
from dashboard_app.bread_crumbs import BreadCrumb, BreadCrumbTrail


def _get_queryset(klass):
    """
    Returns a QuerySet from a Model, Manager, or QuerySet. Created to make
    get_object_or_404 and get_list_or_404 more DRY.
    """
    if isinstance(klass, QuerySet):
        return klass
    elif isinstance(klass, Manager):
        manager = klass
    else:
        manager = klass._default_manager
    return manager.all()


def get_restricted_object_or_404(klass, via, user, *args, **kwargs):
    """
    Uses get() to return an object, or raises a Http404 exception if the object
    does not exist. If the object exists access control check is made
    using the via callback (via is called with the found object and the return
    value must be a RestrictedResource subclass.

    klass may be a Model, Manager, or QuerySet object. All other passed
    arguments and keyword arguments are used in the get() query.

    Note: Like with get(), an MultipleObjectsReturned will be raised if more than one
    object is found.
    """
    queryset = _get_queryset(klass)
    try:
        obj = queryset.get(*args, **kwargs)
        ownership_holder = via(obj)
        if not ownership_holder.is_accessible_by(user):
            raise queryset.model.DoesNotExist()
        return obj
    except queryset.model.DoesNotExist:
        raise Http404('No %s matches the given query.' % queryset.model._meta.object_name)


@BreadCrumb("Dashboard")
def index(request):
    return render_to_response(
        "dashboard_app/index.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(index)
        }, RequestContext(request))


@BreadCrumb("Bundle Streams", parent=index)
def bundle_stream_list(request):
    """
    List of bundle streams.
    """
    return render_to_response(
        'dashboard_app/bundle_stream_list.html', {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                bundle_stream_list),
            "bundle_stream_list": BundleStream.objects.accessible_by_principal(request.user).order_by('pathname'),
            'has_personal_streams': (
                request.user.is_authenticated() and
                BundleStream.objects.filter(user=request.user).count() > 0),
            'has_team_streams': (
                request.user.is_authenticated() and
                BundleStream.objects.filter(
                    group__in = request.user.groups.all()).count() > 0),
        }, RequestContext(request)
    )


@BreadCrumb(
    "Bundles in {pathname}",
    parent=bundle_stream_list,
    needs=['pathname'])
def bundle_list(request, pathname):
    """
    List of bundles in a specified bundle stream.
    """
    bundle_stream = get_restricted_object_or_404(
        BundleStream,
        lambda bundle_stream: bundle_stream,
        request.user,
        pathname=pathname
    )
    return object_list(
        request,
        queryset=bundle_stream.bundles.all().order_by('-uploaded_on'),
        template_name="dashboard_app/bundle_list.html",
        template_object_name="bundle",
        extra_context={
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                bundle_list,
                pathname=pathname),
            "bundle_stream": bundle_stream
        })


@BreadCrumb(
    "Bundle {content_sha1}",
    parent=bundle_list,
    needs=['pathname', 'content_sha1'])
def bundle_detail(request, pathname, content_sha1):
    """
    Detail about a bundle from a particular stream
    """
    bundle_stream = get_restricted_object_or_404(
        BundleStream,
        lambda bundle_stream: bundle_stream,
        request.user,
        pathname=pathname
    )
    return object_detail(
        request,
        queryset=bundle_stream.bundles.all(),
        slug=content_sha1,
        slug_field="content_sha1",
        template_name="dashboard_app/bundle_detail.html",
        template_object_name="bundle",
        extra_context={
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                bundle_detail,
                pathname=pathname,
                content_sha1=content_sha1),
            "bundle_stream": bundle_stream
        })


def bundle_json(request, pathname, content_sha1):
    bundle_stream = get_restricted_object_or_404(
        BundleStream,
        lambda bundle_stream: bundle_stream,
        request.user,
        pathname=pathname
    )
    bundle = bundle_stream.bundles.get(content_sha1=content_sha1)
    test_runs = []
    for test_run in bundle.test_runs.all():
        test_runs.append({
            'name': test_run.test.test_id,
            'url': request.build_absolute_uri(test_run.get_absolute_url()),
            'results': test_run.get_summary_results()
            })
    json_text = json.dumps({
        'test_runs':test_runs,
        })
    content_type = 'application/json'
    if 'callback' in request.GET:
        json_text = '%s(%s)'%(request.GET['callback'], json_text)
        content_type = 'text/javascript'
    return HttpResponse(json_text, content_type=content_type)


def ajax_bundle_viewer(request, pk):
    bundle = get_restricted_object_or_404(
        Bundle,
        lambda bundle: bundle.bundle_stream,
        request.user,
        pk=pk
    )
    return render_to_response(
        "dashboard_app/_ajax_bundle_viewer.html", {
            "bundle": bundle
        },
        RequestContext(request))


@BreadCrumb(
    "Test runs in {pathname}",
    parent=bundle_stream_list,
    needs=['pathname'])
def test_run_list(request, pathname):
    """
    List of test runs in a specified bundle stream.
    """
    bundle_stream = get_restricted_object_or_404(
        BundleStream,
        lambda bundle_stream: bundle_stream,
        request.user,
        pathname=pathname
    )
    return render_to_response(
        'dashboard_app/test_run_list.html', {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                test_run_list,
                pathname=pathname),
            "test_run_list": TestRun.objects.filter(
                bundle__bundle_stream=bundle_stream),
            "bundle_stream": bundle_stream,
        }, RequestContext(request)
    )


@BreadCrumb(
    "Run {analyzer_assigned_uuid}",
    parent=bundle_detail,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid'])
def test_run_detail(request, pathname, content_sha1, analyzer_assigned_uuid):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid
    )
    return render_to_response(
        "dashboard_app/test_run_detail.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                test_run_detail,
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid),
            "test_run": test_run
        }, RequestContext(request))


@BreadCrumb(
    "Software Context",
    parent=test_run_detail,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid'])
def test_run_software_context(request, pathname, content_sha1, analyzer_assigned_uuid):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid
    )
    return render_to_response(
        "dashboard_app/test_run_software_context.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                test_run_software_context,
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid),
            "test_run": test_run
        }, RequestContext(request))


@BreadCrumb(
    "Hardware Context",
    parent=test_run_detail,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid'])
def test_run_hardware_context(request, pathname, content_sha1, analyzer_assigned_uuid):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid
    )
    return render_to_response(
        "dashboard_app/test_run_hardware_context.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                test_run_hardware_context,
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid),
            "test_run": test_run
        }, RequestContext(request))


@BreadCrumb(
    "Details of result {relative_index}",
    parent=test_run_detail,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid', 'relative_index'])
def test_result_detail(request, pathname, content_sha1, analyzer_assigned_uuid, relative_index):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid
    )
    test_result = test_run.test_results.get(relative_index=relative_index)
    return render_to_response(
        "dashboard_app/test_result_detail.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                test_result_detail,
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid,
                relative_index=relative_index),
            "test_result": test_result
        }, RequestContext(request))


@BreadCrumb(
    "Attachments",
    parent=test_run_detail,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid'])
def attachment_list(request, pathname, content_sha1, analyzer_assigned_uuid):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid
    )
    return object_list(
        request,
        queryset=test_run.attachments.all(),
        template_name="dashboard_app/attachment_list.html",
        template_object_name="attachment",
        extra_context={
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                attachment_list, 
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid),
            'test_run': test_run})


@BreadCrumb(
    "{content_filename}",
    parent=attachment_list,
    needs=['pathname', 'content_sha1', 'analyzer_assigned_uuid', 'pk'])
def attachment_detail(request, pathname, content_sha1, analyzer_assigned_uuid, pk):
    attachment = get_restricted_object_or_404(
        Attachment,
        lambda attachment: attachment.test_run.bundle.bundle_stream,
        request.user,
        pk = pk
    )
    return render_to_response(
        "dashboard_app/attachment_detail.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                attachment_detail,
                pathname=pathname,
                content_sha1=content_sha1,
                analyzer_assigned_uuid=analyzer_assigned_uuid,
                pk=pk,
                content_filename=attachment.content_filename),
            "attachment": attachment,
        }, RequestContext(request))


def ajax_attachment_viewer(request, pk):
    attachment = get_restricted_object_or_404(
        Attachment,
        lambda attachment: attachment.test_run.bundle.bundle_stream,
        request.user,
        pk=pk
    )
    if attachment.mime_type == "text/plain":
        data = attachment.get_content_if_possible(mirror=request.user.is_authenticated())
    else:
        data = None
    return render_to_response(
        "dashboard_app/_ajax_attachment_viewer.html", {
            "attachment": attachment,
            "lines": data.splitlines() if data else None,
        },
        RequestContext(request))


@BreadCrumb("Reports", parent=index)
def report_list(request):
    return render_to_response(
        "dashboard_app/report_list.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(report_list),
            "report_list": DataReport.repository.all()
        }, RequestContext(request))


@BreadCrumb("{title}", parent=report_list, needs=['name'])
def report_detail(request, name):
    try:
        report = DataReport.repository.get(name=name)
    except DataReport.DoesNotExist:
        raise Http404('No report matches given name.')
    return render_to_response(
        "dashboard_app/report_detail.html", {
            "is_iframe": request.GET.get("iframe") == "yes",
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                report_detail,
                name=report.name,
                title=report.title),
            "report": report,
        }, RequestContext(request))


@BreadCrumb("Data views", parent=index)
def data_view_list(request):
    return render_to_response(
        "dashboard_app/data_view_list.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(data_view_list),
            "data_view_list": DataView.repository.all(),
        }, RequestContext(request))


@BreadCrumb(
    "Details of {name}",
    parent=data_view_list,
    needs=['name'])
def data_view_detail(request, name):
    try:
        data_view = DataView.repository.get(name=name)
    except DataView.DoesNotExist:
        raise Http404('No data view matches the given query.') 
    return render_to_response(
        "dashboard_app/data_view_detail.html", {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                data_view_detail,
                name=data_view.name,
                summary=data_view.summary),
            "data_view": data_view 
        }, RequestContext(request))


@BreadCrumb("Tests", parent=index)
def test_list(request):
    return object_list(
        request,
        queryset=Test.objects.all(),
        template_name="dashboard_app/test_list.html",
        template_object_name="test",
        extra_context={
            'bread_crumb_trail': BreadCrumbTrail.leading_to(test_list)
        })


@BreadCrumb("Details of {test_id}", parent=test_list, needs=['test_id'])
def test_detail(request, test_id):
    return object_detail(
        request,
        queryset=Test.objects.all(),
        slug=test_id,
        slug_field="test_id",
        template_name="dashboard_app/test_detail.html",
        template_object_name="test",
        extra_context={
            'bread_crumb_trail': BreadCrumbTrail.leading_to(test_detail, test_id=test_id)
        })


def redirect_to(request, object, trailing):
    url = object.get_absolute_url() + trailing
    qs = request.META.get('QUERY_STRING')
    if qs:
        url += '?' + qs
    return redirect(url)


def redirect_to_test_run(request, analyzer_assigned_uuid, trailing=''):
    test_run = get_restricted_object_or_404(
        TestRun,
        lambda test_run: test_run.bundle.bundle_stream,
        request.user,
        analyzer_assigned_uuid=analyzer_assigned_uuid)
    return redirect_to(request, test_run, trailing)


def redirect_to_test_result(request, analyzer_assigned_uuid, relative_index,
                            trailing=''):
    test_result = get_restricted_object_or_404(
        TestResult,
        lambda test_result: test_result.test_run.bundle.bundle_stream,
        request.user,
        test_run__analyzer_assigned_uuid=analyzer_assigned_uuid,
        relative_index=relative_index)
    return redirect_to(request, test_result, trailing)


def redirect_to_bundle(request, content_sha1, trailing=''):
    bundle = get_restricted_object_or_404(
        Bundle,
        lambda bundle: bundle.bundle_stream,
        request.user,
        content_sha1=content_sha1)
    return redirect_to(request, bundle, trailing)


@BreadCrumb("Image Status Matrix", parent=index)
def image_status_list(request):
    return render_to_response(
        "dashboard_app/image_status_list.html", {
            'hwpack_list': ImageHealth.get_hwpack_list(),
            'rootfs_list': ImageHealth.get_rootfs_list(),
            'ImageHealth': ImageHealth,
            'bread_crumb_trail': BreadCrumbTrail.leading_to(image_status_list)
        }, RequestContext(request))


@BreadCrumb(
    "Image Status for {rootfs_type} + {hwpack_type}",
    parent=image_status_list,
    needs=["rootfs_type", "hwpack_type"])
def image_status_detail(request, rootfs_type, hwpack_type):
    image_health = ImageHealth(rootfs_type, hwpack_type)
    return render_to_response(
        "dashboard_app/image_status_detail.html", {
            'image_health': image_health,
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                image_status_detail,
                rootfs_type=rootfs_type,
                hwpack_type=hwpack_type),
        }, RequestContext(request))


@BreadCrumb(
    "Test history for {test_id}",
    parent=image_status_detail,
    needs=["rootfs_type", "hwpack_type", "test_id"])
def image_test_history(request, rootfs_type, hwpack_type, test_id):
    image_health = ImageHealth(rootfs_type, hwpack_type)
    test = get_object_or_404(Test, test_id=test_id)
    test_run_list = image_health.get_test_runs().filter(test=test)
    return render_to_response(
        "dashboard_app/image_test_history.html", {
            'test_run_list': test_run_list,
            'test': test,
            'image_health': image_health,
            'bread_crumb_trail': BreadCrumbTrail.leading_to(
                image_test_history,
                rootfs_type=rootfs_type,
                hwpack_type=hwpack_type,
                test=test,
                test_id=test_id),
        }, RequestContext(request))



# Copyright (C) 2010 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of Launch Control.
#
# Launch Control is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# Launch Control is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dashboard_app.views.bundle_stream_list
"""

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse

from dashboard_app.tests import fixtures
from dashboard_app.tests.utils import (
    DashboardViewsTestCase,
    TestClient,
)


class BundleStreamListViewAnonymousTest(DashboardViewsTestCase):

    scenarios = [
        ('empty', {
            'bundle_streams': [],
        }),
        ('public_streams', {
            'bundle_streams': [
                '/anonymous/',
                '/anonymous/name/',
                '/public/personal/user/',
                '/public/personal/user/name/',
                '/public/team/group/',
                '/public/team/group/name/',
            ]
        }),
        ('private_streams', {
            'bundle_streams': [
                '/private/personal/user/',
                '/private/personal/user/name/',
                '/private/team/group/',
                '/private/team/group/name/',
            ]
        }),
    ]

    def setUp(self):
        super(BundleStreamListViewAnonymousTest, self).setUp()
        self.url = reverse("dashboard_app.views.bundle_stream_list")
        self.user = None

    def test_status_code(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

    def test_template_used(self):
        response = self.client.get(self.url)
        self.assertTemplateUsed(response,
                "dashboard_app/bundle_stream_list.html")

    def test_listed_bundles_are_the_ones_we_should_see(self):
        with fixtures.created_bundle_streams(self.bundle_streams) as bundle_streams:
            response = self.client.get(self.url)
            expected_bsl = sorted(
                    [bundle_stream.pk for bundle_stream in
                        bundle_streams if
                        bundle_stream.is_accessible_by(self.user)])
            effective_bsl = sorted(
                    [bundle_stream.pk for bundle_stream in
                        response.context['bundle_stream_list']])
            self.assertEqual(effective_bsl, expected_bsl)


class BundleStreamListViewAuthorizedTest(BundleStreamListViewAnonymousTest):

    def setUp(self):
        super(BundleStreamListViewAuthorizedTest, self).setUp()
        self.client = TestClient()
        self.user = User.objects.create(username='user')
        self.user.groups.create(name='group')
        self.client.login_user(self.user)
